# Copyright 2019 Canonical, Ltd.
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as
# published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import unittest
from unittest import mock

import urwid

from subiquitycore.testing import view_helpers
from subiquitycore.view import BaseView

from subiquity.controllers.filesystem import FilesystemController
from subiquity.ui.views.filesystem.lvm import VolGroupStretchy
from subiquity.ui.views.filesystem.tests.test_partition import (
    make_model_and_disk,
    )


def make_view(model, existing=None):
    controller = mock.create_autospec(spec=FilesystemController)
    base_view = BaseView(urwid.Text(""))
    base_view.model = model
    base_view.controller = controller
    base_view.refresh_model_inputs = lambda: None
    stretchy = VolGroupStretchy(base_view, existing)
    base_view.show_stretchy_overlay(stretchy)
    return base_view, stretchy


class LVMViewTests(unittest.TestCase):

    def test_create_vg(self):
        model, disk = make_model_and_disk()
        part1 = model.add_partition(disk, 10*(2**30))
        part2 = model.add_partition(disk, 10*(2**30))
        view, stretchy = make_view(model)
        form_data = {
            'name': 'vg1',
            'devices': {part1: 'active', part2: 'active'},
            }
        expected_data = {
            'name': 'vg1',
            'devices': {part1, part2},
            'encrypt': False,
            }
        view_helpers.enter_data(stretchy.form, form_data)
        view_helpers.click(stretchy.form.done_btn.base_widget)
        view.controller.volgroup_handler.assert_called_once_with(
            None, expected_data)

    def test_create_vg_encrypted(self):
        model, disk = make_model_and_disk()
        part1 = model.add_partition(disk, 10*(2**30))
        part2 = model.add_partition(disk, 10*(2**30))
        view, stretchy = make_view(model)
        form_data = {
            'name': 'vg1',
            'devices': {part1: 'active', part2: 'active'},
            'encrypt': True,
            'password': 'passw0rd',
            'confirm_password': 'passw0rd',
            }
        expected_data = {
            'name': 'vg1',
            'devices': {part1, part2},
            'encrypt': True,
            'password': 'passw0rd',
            }
        view_helpers.enter_data(stretchy.form, form_data)
        view_helpers.click(stretchy.form.done_btn.base_widget)
        view.controller.volgroup_handler.assert_called_once_with(
            None, expected_data)



# Copyright 2018 Canonical, Ltd.
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as
# published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import attr

from urwid import (
    ACTIVATE,
    AttrWrap,
    Button,
    connect_signal,
    LineBox,
    PopUpLauncher,
    SelectableIcon,
    Text,
    Widget,
    )

from subiquitycore.ui.container import (
    Columns,
    ListBox,
    WidgetWrap,
)
from subiquitycore.ui.utils import Color


class ActionBackButton(Button):
    button_left = Text("\N{BLACK LEFT-POINTING SMALL TRIANGLE}")
    button_right = Text("")


class ActionMenuOpenButton(Button):
    button_left = Text("")
    button_right = Text("\N{BLACK RIGHT-POINTING SMALL TRIANGLE}")


class ActionMenuButton(Button):
    button_left = Text("")
    button_right = Text("")


class _ActionMenuDialog(WidgetWrap):
    """A list of action buttons with a box around them."""

    def __init__(self, parent):
        self.parent = parent
        close_text = "(close)"
        close = ActionBackButton(close_text)
        connect_signal(close, "click", self.close)
        group = [Color.menu_button(close)]
        width = len(close_text)
        for i, action in enumerate(self.parent._actions):
            if action.enabled:
                if isinstance(action.label, Widget):
                    btn = action.label
                elif action.opens_dialog:
                    btn = Color.menu_button(ActionMenuOpenButton(action.label))
                else:
                    btn = Color.menu_button(ActionMenuButton(action.label))
                width = max(width, len(btn.base_widget.label))
                connect_signal(
                    btn.base_widget, 'click', self.click, action.value)
            else:
                label = action.label
                if isinstance(label, Widget):
                    label = label.base_widget.label
                width = max(width, len(label))
                if action.opens_dialog:
                    rhs = "\N{BLACK RIGHT-POINTING SMALL TRIANGLE}"
                else:
                    rhs = ""
                btn = Columns([
                    ('fixed', 1, Text("")),
                    Text(label),
                    ('fixed', 1, Text(rhs)),
                    ], dividechars=1)
                btn = AttrWrap(btn, 'info_minor')
            group.append(btn)
        self.width = width
        super().__init__(Color.body(LineBox(ListBox(group))))

    def close(self, sender):
        self.parent.close_pop_up()

    def click(self, btn, value):
        self.parent._action(value)
        self.parent.close_pop_up()

    def keypress(self, size, key):
        if key == 'esc':
            self.parent.close_pop_up()
        else:
            return super().keypress(size, key)


@attr.s
class Action:
    # The label that is shown in the menu
    label = attr.ib()
    enabled = attr.ib()
    # The value passed along with the 'action' signal
    value = attr.ib()
    # Actions that open a dialog get a > at the end.
    opens_dialog = attr.ib(default=False)


class ActionMenu(PopUpLauncher):

    signals = ['action', 'open', 'close']

    def __init__(self, opts,
                 icon="\N{BLACK RIGHT-POINTING SMALL TRIANGLE}"):
        self._actions = []
        for opt in opts:
            if not isinstance(opt, Action):
                opt = Action(*opt)
            self._actions.append(opt)
        self.icon = icon
        self._button = SelectableIcon(self.icon, 0)
        super().__init__(self._button)
        self._dialog = _ActionMenuDialog(self)

    def get_natural_width(self):
        return len(self.icon)

    def keypress(self, size, key):
        if self._command_map[key] != ACTIVATE:
            return key
        self.open_pop_up()

    def _action(self, action):
        self._emit("action", action)

    def open_pop_up(self):
        self._dialog._w.base_widget.focus_position = 0
        self._emit("open")
        super().open_pop_up()

    def close_pop_up(self):
        self._emit("close")
        super().close_pop_up()

    def create_pop_up(self):
        return self._dialog

    def get_pop_up_parameters(self):
        width = self._dialog.width + 7
        return {
            'left': 1,
            'top': -1,
            'overlay_width': width,
            'overlay_height': len(self._actions) + 3,
            }



#!/usr/bin/env python3
# Copyright 2015 Canonical, Ltd.
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as
# published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import argparse
import logging
import os
import fcntl
import sys
import time

from cloudinit import atomic_helper, safeyaml, stages

from subiquitycore.log import setup_logger
from subiquitycore.utils import run_command


class ClickAction(argparse.Action):
    def __call__(self, parser, namespace, values, option_string=None):
        namespace.scripts.append("c(" + repr(values) + ")")


def parse_options(argv):
    parser = argparse.ArgumentParser(
        description='SUbiquity - Ubiquity for Servers',
        prog='subiquity')
    try:
        ascii_default = os.ttyname(0) == "/dev/ttysclp0"
    except OSError:
        ascii_default = False
    parser.set_defaults(ascii=ascii_default)
    parser.add_argument('--dry-run', action='store_true',
                        dest='dry_run',
                        help='menu-only, do not call installer function')
    parser.add_argument('--serial', action='store_true',
                        dest='run_on_serial',
                        help='Run the installer over serial console.')
    parser.add_argument('--ssh', action='store_true',
                        dest='ssh',
                        help='Print ssh login details')
    parser.add_argument('--ascii', action='store_true',
                        dest='ascii',
                        help='Run the installer in ascii mode.')
    parser.add_argument('--unicode', action='store_false',
                        dest='ascii',
                        help='Run the installer in unicode mode.')
    parser.add_argument('--machine-config', metavar='CONFIG',
                        dest='machine_config',
                        help="Don't Probe. Use probe data file")
    parser.add_argument('--bootloader',
                        choices=['none', 'bios', 'prep', 'uefi'],
                        help='Override style of bootloader to use')
    parser.add_argument('--screens', action='append', dest='screens',
                        default=[])
    parser.add_argument('--script', metavar="SCRIPT", action='append',
                        dest='scripts', default=[],
                        help=('Execute SCRIPT in a namespace containing view '
                              'helpers and "ui"'))
    parser.add_argument('--click', metavar="PAT", action=ClickAction,
                        help='Synthesize a click on a button matching PAT')
    parser.add_argument('--answers')
    parser.add_argument('--autoinstall', action='store')
    with open('/proc/cmdline') as fp:
        cmdline = fp.read()
    parser.add_argument('--kernel-cmdline', action='store', default=cmdline)
    parser.add_argument('--source', default=[], action='append',
                        dest='sources', metavar='URL',
                        help='install from url instead of default.')
    parser.add_argument(
        '--snaps-from-examples', action='store_const', const=True,
        dest="snaps_from_examples",
        help=("Load snap details from examples/snaps instead of store. "
              "Default in dry-run mode.  "
              "See examples/snaps/README.md for more."))
    parser.add_argument(
        '--no-snaps-from-examples', action='store_const', const=False,
        dest="snaps_from_examples",
        help=("Load snap details from store instead of examples. "
              "Default in when not in dry-run mode.  "
              "See examples/snaps/README.md for more."))
    parser.add_argument(
        '--snap-section', action='store', default='server',
        help=("Show snaps from this section of the store in the snap "
              "list screen."))
    return parser.parse_args(argv)


LOGDIR = "/var/log/installer/"

AUTO_ANSWERS_FILE = "/subiquity_config/answers.yaml"


def main():
    # Prefer utils from $SNAP, over system-wide
    snap = os.environ.get('SNAP')
    if snap:
        os.environ['PATH'] = os.pathsep.join([
            os.path.join(snap, 'bin'),
            os.path.join(snap, 'usr', 'bin'),
            os.environ['PATH'],
        ])
        os.environ["APPORT_DATA_DIR"] = os.path.join(snap, 'share/apport')
    # This must come after setting $APPORT_DATA_DIR.
    from subiquity.core import Subiquity
    opts = parse_options(sys.argv[1:])
    global LOGDIR
    if opts.dry_run:
        LOGDIR = ".subiquity"
        if opts.snaps_from_examples is None:
            opts.snaps_from_examples = True
    logfiles = setup_logger(dir=LOGDIR)

    logger = logging.getLogger('subiquity')
    version = os.environ.get("SNAP_REVISION", "unknown")
    logger.info("Starting Subiquity revision {}".format(version))
    logger.info("Arguments passed: {}".format(sys.argv))

    if not opts.dry_run:
        ci_start = time.time()
        status_txt = run_command(["cloud-init", "status", "--wait"]).stdout
        logger.debug("waited %ss for cloud-init", time.time() - ci_start)
        if "status: done" in status_txt:
            logger.debug("loading cloud config")
            init = stages.Init()
            init.read_cfg()
            init.fetch(existing="trust")
            cloud = init.cloudify()
            autoinstall_path = '/autoinstall.yaml'
            if 'autoinstall' in cloud.cfg:
                if not os.path.exists(autoinstall_path):
                    atomic_helper.write_file(
                        autoinstall_path,
                        safeyaml.dumps(
                            cloud.cfg['autoinstall']).encode('utf-8'),
                        mode=0o600)
            if os.path.exists(autoinstall_path):
                opts.autoinstall = autoinstall_path
        else:
            logger.debug(
                "cloud-init status: %r, assumed disabled",
                status_txt)

    block_log_dir = os.path.join(LOGDIR, "block")
    os.makedirs(block_log_dir, exist_ok=True)
    handler = logging.FileHandler(os.path.join(block_log_dir, 'discover.log'))
    handler.setLevel('DEBUG')
    handler.setFormatter(
        logging.Formatter("%(asctime)s %(name)s:%(lineno)d %(message)s"))
    logging.getLogger('probert').addHandler(handler)
    handler.addFilter(lambda rec: rec.name != 'probert.network')
    logging.getLogger('curtin').addHandler(handler)
    logging.getLogger('block-discover').addHandler(handler)

    if opts.ssh:
        from subiquity.ui.views.help import (
            ssh_help_texts, get_installer_password)
        from subiquitycore.ssh import get_ips_standalone
        texts = ssh_help_texts(
            get_ips_standalone(), get_installer_password(opts.dry_run))
        for line in texts:
            if hasattr(line, 'text'):
                if line.text.startswith('installer@'):
                    print(' ' * 4 + line.text)
                else:
                    print(line.text)
            else:
                print(line)
        return 0

    if opts.answers is None and os.path.exists(AUTO_ANSWERS_FILE):
        logger.debug("Autoloading answers from %s", AUTO_ANSWERS_FILE)
        opts.answers = AUTO_ANSWERS_FILE

    if opts.answers:
        opts.answers = open(opts.answers)
        try:
            fcntl.flock(opts.answers, fcntl.LOCK_EX | fcntl.LOCK_NB)
        except OSError:
            logger.exception(
                'Failed to lock auto answers file, proceding without it.')
            opts.answers.close()
            opts.answers = None

    subiquity_interface = Subiquity(opts, block_log_dir)

    subiquity_interface.note_file_for_apport(
        "InstallerLog", logfiles['debug'])
    subiquity_interface.note_file_for_apport(
        "InstallerLogInfo", logfiles['info'])

    subiquity_interface.run()


if __name__ == '__main__':
    sys.exit(main())



# Copyright 2019 Canonical, Ltd.
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as
# published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import copy
import logging

from curtin.reporter import (
    available_handlers,
    update_configuration,
    )
from curtin.reporter.events import (
    report_finish_event,
    report_start_event,
    status,
    )
from curtin.reporter.handlers import (
    LogHandler,
    )

from subiquity.controller import NoUIController


class LogHandler(LogHandler):
    def publish_event(self, event):
        level = getattr(logging, event.level)
        logger = logging.getLogger('')
        logger.log(level, event.as_string())


available_handlers.unregister_item('log')
available_handlers.register_item('log', LogHandler)

INITIAL_CONFIG = {'logging': {'type': 'log'}}
NON_INTERACTIVE_CONFIG = {'builtin': {'type': 'print'}}


class ReportingController(NoUIController):

    autoinstall_key = "reporting"
    autoinstall_schema = {
        'type': 'object',
        'additionalProperties': {
            'type': 'object',
            'properties': {
                'type': {'type': 'string'},
                },
            'required': ['type'],
            'additionalProperties': True,
            },
        }

    def __init__(self, app):
        self.config = copy.deepcopy(INITIAL_CONFIG)
        super().__init__(app)

    def load_autoinstall_data(self, data):
        if self.app.interactive():
            return
        self.config.update(copy.deepcopy(NON_INTERACTIVE_CONFIG))
        if data is not None:
            self.config.update(copy.deepcopy(data))

    def start(self):
        update_configuration(self.config)

    def report_start_event(self, name, description, level):
        report_start_event(name, description, level=level)

    def report_finish_event(self, name, description, result, level):
        result = getattr(status, result.name, status.WARN)
        report_finish_event(name, description, result, level=level)



# Copyright 2015 Canonical, Ltd.
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as
# published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import logging
from urwid import (
    LineBox,
    Text,
    )

from subiquitycore.view import BaseView
from subiquitycore.ui.buttons import (
    cancel_btn,
    danger_btn,
    ok_btn,
    other_btn,
    )
from subiquitycore.ui.container import Columns, ListBox, Pile
from subiquitycore.ui.form import Toggleable
from subiquitycore.ui.spinner import Spinner
from subiquitycore.ui.utils import button_pile, Padding, rewrap
from subiquitycore.ui.stretchy import Stretchy
from subiquitycore.ui.width import widget_width

log = logging.getLogger("subiquity.views.installprogress")


class MyLineBox(LineBox):
    def format_title(self, title):
        if title:
            return [" ", title, " "]
        else:
            return ""


class ProgressView(BaseView):

    title = _("Install progress")

    def __init__(self, controller):
        self.controller = controller
        self.ongoing = {}  # context -> line containing a spinner

        self.reboot_btn = Toggleable(ok_btn(
            _("Reboot Now"), on_press=self.reboot))
        self.view_error_btn = cancel_btn(
            _("View error report"), on_press=self.view_error)
        self.view_log_btn = other_btn(
            _("View full log"), on_press=self.view_log)
        self.continue_btn = other_btn(
            _("Continue"), on_press=self.continue_)

        self.event_listbox = ListBox()
        self.event_linebox = MyLineBox(self.event_listbox)
        self.event_buttons = button_pile([self.view_log_btn])
        event_body = [
            ('weight', 1, Padding.center_79(self.event_linebox, min_width=76)),
            ('pack', Text("")),
            ('pack', self.event_buttons),
            ('pack', Text("")),
        ]
        self.event_pile = Pile(event_body)

        self.log_listbox = ListBox()
        log_linebox = MyLineBox(self.log_listbox, _("Full installer output"))
        log_body = [
            ('weight', 1, log_linebox),
            ('pack', button_pile([other_btn(_("Close"),
                                  on_press=self.close_log)])),
            ]
        self.log_pile = Pile(log_body)

        super().__init__(self.event_pile)

    def _add_line(self, lb, line):
        lb = lb.base_widget
        walker = lb.body
        at_end = len(walker) == 0 or lb.focus_position == len(walker) - 1
        walker.append(line)
        if at_end:
            lb.set_focus(len(walker) - 1)
            lb.set_focus_valign('bottom')

    def event_start(self, context, message):
        self.event_finish(context.parent)
        walker = self.event_listbox.base_widget.body
        indent = '  ' * (context.full_name().count('/') - 2)
        spinner = Spinner(self.controller.app.aio_loop)
        spinner.start()
        new_line = Columns([
            ('pack', Text(indent + message)),
            ('pack', spinner),
            ], dividechars=1)
        self.ongoing[context] = len(walker)
        self._add_line(self.event_listbox, new_line)

    def event_finish(self, context):
        index = self.ongoing.pop(context, None)
        if index is None:
            return
        walker = self.event_listbox.base_widget.body
        spinner = walker[index][1]
        spinner.stop()
        walker[index] = walker[index][0]

    def finish_all(self):
        for context in self.ongoing.copy():
            self.event_finish(context)

    def add_log_line(self, text):
        self._add_line(self.log_listbox, Text(text))

    def set_status(self, text):
        self.event_linebox.set_title(text)

    def _set_button_width(self):
        w = 14
        for b, o in self.event_buttons.original_widget.contents:
            w = max(widget_width(b), w)
        self.event_buttons.width = self.event_buttons.min_width = w

    def _set_buttons(self, buttons):
        p = self.event_buttons.original_widget
        p.contents[:] = [(b, p.options('pack')) for b in buttons]
        self._set_button_width()

    def update_running(self):
        self.reboot_btn.base_widget.set_label(_("Cancel update and reboot"))
        self._set_button_width()

    def update_done(self):
        self.reboot_btn.base_widget.set_label(_("Reboot"))
        self._set_button_width()

    def show_complete(self):
        btns = [self.view_log_btn, self.reboot_btn]
        self._set_buttons(btns)
        self.event_buttons.base_widget.focus_position = 1
        self.event_pile.base_widget.focus_position = 2

    def show_continue(self):
        btns = [self.continue_btn, self.reboot_btn]
        self._set_buttons(btns)
        self.event_buttons.base_widget.focus_position = 0
        self.event_pile.base_widget.focus_position = 2

    def continue_(self, sender=None):
        self.controller.app.next_screen()

    def hide_continue(self):
        btns = [self.view_log_btn]
        self._set_buttons(btns)
        self.event_buttons.base_widget.focus_position = 0
        self.event_pile.base_widget.focus_position = 2

    def show_error(self, crash_report):
        btns = [self.view_log_btn, self.view_error_btn, self.reboot_btn]
        self._set_buttons(btns)
        self.event_buttons.base_widget.focus_position = 1
        self.event_pile.base_widget.focus_position = 2
        self.crash_report = crash_report
        self.controller.app.show_error_report(crash_report)

    def reboot(self, btn):
        self.reboot_btn.base_widget.set_label(_("Rebooting..."))
        self.reboot_btn.enabled = False
        self.event_buttons.original_widget._select_first_selectable()
        self.controller.click_reboot()
        self._set_button_width()

    def view_error(self, btn):
        self.controller.app.show_error_report(self.crash_report)

    def view_log(self, btn):
        self._w = self.log_pile

    def close_log(self, btn):
        self._w = self.event_pile


confirmation_text = _("""\
Selecting Continue below will begin the installation process and
result in the loss of data on the disks selected to be formatted.

You will not be able to return to this or a previous screen once
the installation has started.

Are you sure you want to continue?""")


class InstallConfirmation(Stretchy):
    def __init__(self, parent, app):
        self.parent = parent
        self.app = app
        widgets = [
            Text(rewrap(_(confirmation_text))),
            Text(""),
            button_pile([
                cancel_btn(_("No"), on_press=self.cancel),
                danger_btn(_("Continue"), on_press=self.ok)]),
            ]
        super().__init__(
            _("Confirm destructive action"),
            widgets,
            stretchy_index=0,
            focus_index=2)

    def ok(self, sender):
        self.app.confirm_install()
        self.parent.remove_overlay()
        if isinstance(self.parent, ProgressView):
            self.parent.hide_continue()
        self.app.next_screen()

    def cancel(self, sender):
        self.parent.remove_overlay()
        if isinstance(self.parent, ProgressView):
            self.parent.show_continue()



# Copyright 2020 Canonical, Ltd.
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as
# published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
import unittest

from console_conf.models.systems import (
    SystemAction,
    )
from console_conf.ui.views.chooser import (
    by_preferred_action_type,
    )


class TestSorting(unittest.TestCase):

    def test_simple(self):
        data = [
            SystemAction(title="b-other", mode="unknown"),
            SystemAction(title="reinstall", mode="install"),
            SystemAction(title="run normally", mode="run"),
            SystemAction(title="a-other", mode="some-other"),
            SystemAction(title="recover", mode="recover"),
        ]
        exp = [
            SystemAction(title="run normally", mode="run"),
            SystemAction(title="recover", mode="recover"),
            SystemAction(title="reinstall", mode="install"),
            SystemAction(title="a-other", mode="some-other"),
            SystemAction(title="b-other", mode="unknown"),
        ]
        self.assertSequenceEqual(sorted(data, key=by_preferred_action_type),
                                 exp)



#!/usr/bin/env python3
# Copyright 2015 Canonical, Ltd.
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as
# published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import argparse
import sys
import os
import logging
from subiquitycore.log import setup_logger
from subiquitycore import __version__ as VERSION
from console_conf.core import ConsoleConf, RecoveryChooser


class ClickAction(argparse.Action):
    def __call__(self, parser, namespace, values, option_string=None):
        namespace.scripts.append("c(" + repr(values) + ")")


def parse_options(argv):
    parser = argparse.ArgumentParser(
        description=(
            'console-conf - Pre-Ownership Configuration for Ubuntu Core'),
        prog='console-conf')
    parser.add_argument('--dry-run', action='store_true',
                        dest='dry_run',
                        help='menu-only, do not call installer function')
    parser.add_argument('--serial', action='store_true',
                        dest='run_on_serial',
                        help='Run the installer over serial console.')
    parser.add_argument('--ascii', action='store_true',
                        dest='ascii',
                        help='Run the installer in ascii mode.')
    parser.add_argument('--machine-config', metavar='CONFIG',
                        dest='machine_config',
                        help="Don't Probe. Use probe data file")
    parser.add_argument('--screens', action='append', dest='screens',
                        default=[])
    parser.add_argument('--script', metavar="SCRIPT", action='append',
                        dest='scripts', default=[],
                        help=('Execute SCRIPT in a namespace containing view '
                              'helpers and "ui"'))
    parser.add_argument('--click', metavar="PAT", action=ClickAction,
                        help='Synthesize a click on a button matching PAT')
    parser.add_argument('--answers')
    parser.add_argument('--recovery-chooser-mode', action='store_true',
                        dest='chooser_systems',
                        help=('Run as a recovery chooser interacting with the '
                              'calling process over stdin/stdout streams'))
    return parser.parse_args(argv)


LOGDIR = "/var/log/console-conf/"


def main():
    opts = parse_options(sys.argv[1:])
    global LOGDIR
    if opts.dry_run:
        LOGDIR = ".subiquity"
    setup_logger(dir=LOGDIR)
    logger = logging.getLogger('console_conf')
    logger.info("Starting console-conf v{}".format(VERSION))
    logger.info("Arguments passed: {}".format(sys.argv))

    if opts.chooser_systems:
        # when running as a chooser, the stdin/stdout streams are set up by the
        # process that runs us, attempt to restore the tty in/out by looking at
        # stderr
        chooser_input, chooser_output = restore_std_streams_from(sys.stderr)
        interface = RecoveryChooser(opts, chooser_input, chooser_output)
    else:
        interface = ConsoleConf(opts)

    interface.run()


def restore_std_streams_from(from_file):
    """
    Attempt to restore the original sys.std{in,out} streams by inspecting the
    tty that stderr is hooked up to. Returns the chooser input/output streams.
    """
    tty = os.ttyname(from_file.fileno())
    # we have tty now
    chooser_input, chooser_output = sys.stdin, sys.stdout
    sys.stdin = open(tty, 'r')
    sys.stdout = open(tty, 'w')
    return chooser_input, chooser_output


if __name__ == '__main__':
    sys.exit(main())



# Copyright 2019 Canonical, Ltd.
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as
# published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import attr
import logging
import platform
import shlex

from collections import OrderedDict
from urwid import Text

from subiquitycore.ui.utils import Color
from subiquitycore.utils import run_command

from subiquity.controller import SubiquityController
from subiquity.ui.views import ZdevView


log = logging.getLogger("subiquitycore.controller.zdev")

lszdev_cmd = ['lszdev', '--pairs', '--columns',
              'id,type,on,exists,pers,auto,failed,names']

lszdev_stock = '''id="0.0.1500" type="dasd-eckd" on="no" exists="yes" pers="no" auto="no" failed="no" names=""
id="0.0.1501" type="dasd-eckd" on="no" exists="yes" pers="no" auto="no" failed="no" names=""
id="0.0.1502" type="dasd-eckd" on="no" exists="yes" pers="no" auto="no" failed="no" names=""
id="0.0.1503" type="dasd-eckd" on="no" exists="yes" pers="no" auto="no" failed="no" names=""
id="0.0.1504" type="dasd-eckd" on="no" exists="yes" pers="no" auto="no" failed="no" names=""
id="0.0.1505" type="dasd-eckd" on="no" exists="yes" pers="no" auto="no" failed="no" names=""
id="0.0.1506" type="dasd-eckd" on="no" exists="yes" pers="no" auto="no" failed="no" names=""
id="0.0.1507" type="dasd-eckd" on="no" exists="yes" pers="no" auto="no" failed="no" names=""
id="0.0.1508" type="dasd-eckd" on="no" exists="yes" pers="no" auto="no" failed="no" names=""
id="0.0.1509" type="dasd-eckd" on="no" exists="yes" pers="no" auto="no" failed="no" names=""
id="0.0.150a" type="dasd-eckd" on="no" exists="yes" pers="no" auto="no" failed="no" names=""
id="0.0.150b" type="dasd-eckd" on="no" exists="yes" pers="no" auto="no" failed="no" names=""
id="0.0.150c" type="dasd-eckd" on="no" exists="yes" pers="no" auto="no" failed="no" names=""
id="0.0.150d" type="dasd-eckd" on="no" exists="yes" pers="no" auto="no" failed="no" names=""
id="0.0.150e" type="dasd-eckd" on="no" exists="yes" pers="no" auto="no" failed="no" names=""
id="0.0.150f" type="dasd-eckd" on="no" exists="yes" pers="no" auto="no" failed="no" names=""
id="0.0.1510" type="dasd-eckd" on="no" exists="yes" pers="no" auto="no" failed="no" names=""
id="0.0.1511" type="dasd-eckd" on="no" exists="yes" pers="no" auto="no" failed="no" names=""
id="0.0.1512" type="dasd-eckd" on="no" exists="yes" pers="no" auto="no" failed="no" names=""
id="0.0.1513" type="dasd-eckd" on="no" exists="yes" pers="no" auto="no" failed="no" names=""
id="0.0.1514" type="dasd-eckd" on="no" exists="yes" pers="no" auto="no" failed="no" names=""
id="0.0.1515" type="dasd-eckd" on="no" exists="yes" pers="no" auto="no" failed="no" names=""
id="0.0.1516" type="dasd-eckd" on="no" exists="yes" pers="no" auto="no" failed="no" names=""
id="0.0.1517" type="dasd-eckd" on="no" exists="yes" pers="no" auto="no" failed="no" names=""
id="0.0.1518" type="dasd-eckd" on="no" exists="yes" pers="no" auto="no" failed="no" names=""
id="0.0.1519" type="dasd-eckd" on="no" exists="yes" pers="no" auto="no" failed="no" names=""
id="0.0.151a" type="dasd-eckd" on="no" exists="yes" pers="no" auto="no" failed="no" names=""
id="0.0.151b" type="dasd-eckd" on="no" exists="yes" pers="no" auto="no" failed="no" names=""
id="0.0.151c" type="dasd-eckd" on="no" exists="yes" pers="no" auto="no" failed="no" names=""
id="0.0.151d" type="dasd-eckd" on="no" exists="yes" pers="no" auto="no" failed="no" names=""
id="0.0.151e" type="dasd-eckd" on="no" exists="yes" pers="no" auto="no" failed="no" names=""
id="0.0.151f" type="dasd-eckd" on="no" exists="yes" pers="no" auto="no" failed="no" names=""
id="0.0.1520" type="dasd-eckd" on="no" exists="yes" pers="no" auto="no" failed="no" names=""
id="0.0.1521" type="dasd-eckd" on="no" exists="yes" pers="no" auto="no" failed="no" names=""
id="0.0.1522" type="dasd-eckd" on="no" exists="yes" pers="no" auto="no" failed="no" names=""
id="0.0.1523" type="dasd-eckd" on="no" exists="yes" pers="no" auto="no" failed="no" names=""
id="0.0.1524" type="dasd-eckd" on="no" exists="yes" pers="no" auto="no" failed="no" names=""
id="0.0.1525" type="dasd-eckd" on="no" exists="yes" pers="no" auto="no" failed="no" names=""
id="0.0.1526" type="dasd-eckd" on="no" exists="yes" pers="no" auto="no" failed="no" names=""
id="0.0.1527" type="dasd-eckd" on="no" exists="yes" pers="no" auto="no" failed="no" names=""
id="0.0.1528" type="dasd-eckd" on="no" exists="yes" pers="no" auto="no" failed="no" names=""
id="0.0.1529" type="dasd-eckd" on="no" exists="yes" pers="no" auto="no" failed="no" names=""
id="0.0.152a" type="dasd-eckd" on="no" exists="yes" pers="no" auto="no" failed="no" names=""
id="0.0.152b" type="dasd-eckd" on="no" exists="yes" pers="no" auto="no" failed="no" names=""
id="0.0.152c" type="dasd-eckd" on="no" exists="yes" pers="no" auto="no" failed="no" names=""
id="0.0.152d" type="dasd-eckd" on="no" exists="yes" pers="no" auto="no" failed="no" names=""
id="0.0.152e" type="dasd-eckd" on="no" exists="yes" pers="no" auto="no" failed="no" names=""
id="0.0.152f" type="dasd-eckd" on="no" exists="yes" pers="no" auto="no" failed="no" names=""
id="0.0.1530" type="dasd-eckd" on="no" exists="yes" pers="no" auto="no" failed="no" names=""
id="0.0.1531" type="dasd-eckd" on="no" exists="yes" pers="no" auto="no" failed="no" names=""
id="0.0.1532" type="dasd-eckd" on="no" exists="yes" pers="no" auto="no" failed="no" names=""
id="0.0.1533" type="dasd-eckd" on="no" exists="yes" pers="no" auto="no" failed="no" names=""
id="0.0.1534" type="dasd-eckd" on="no" exists="yes" pers="no" auto="no" failed="no" names=""
id="0.0.1535" type="dasd-eckd" on="no" exists="yes" pers="no" auto="no" failed="no" names=""
id="0.0.1536" type="dasd-eckd" on="no" exists="yes" pers="no" auto="no" failed="no" names=""
id="0.0.1537" type="dasd-eckd" on="no" exists="yes" pers="no" auto="no" failed="no" names=""
id="0.0.1538" type="dasd-eckd" on="no" exists="yes" pers="no" auto="no" failed="no" names=""
id="0.0.1539" type="dasd-eckd" on="no" exists="yes" pers="no" auto="no" failed="no" names=""
id="0.0.153a" type="dasd-eckd" on="no" exists="yes" pers="no" auto="no" failed="no" names=""
id="0.0.153b" type="dasd-eckd" on="no" exists="yes" pers="no" auto="no" failed="no" names=""
id="0.0.153c" type="dasd-eckd" on="no" exists="yes" pers="no" auto="no" failed="no" names=""
id="0.0.153d" type="dasd-eckd" on="no" exists="yes" pers="no" auto="no" failed="no" names=""
id="0.0.153e" type="dasd-eckd" on="no" exists="yes" pers="no" auto="no" failed="no" names=""
id="0.0.153f" type="dasd-eckd" on="no" exists="yes" pers="no" auto="no" failed="no" names=""
id="0.0.1540" type="dasd-eckd" on="no" exists="yes" pers="no" auto="no" failed="no" names=""
id="0.0.1541" type="dasd-eckd" on="no" exists="yes" pers="no" auto="no" failed="no" names=""
id="0.0.1542" type="dasd-eckd" on="no" exists="yes" pers="no" auto="no" failed="no" names=""
id="0.0.1543" type="dasd-eckd" on="no" exists="yes" pers="no" auto="no" failed="no" names=""
id="0.0.1544" type="dasd-eckd" on="no" exists="yes" pers="no" auto="no" failed="no" names=""
id="0.0.1545" type="dasd-eckd" on="no" exists="yes" pers="no" auto="no" failed="no" names=""
id="0.0.1546" type="dasd-eckd" on="no" exists="yes" pers="no" auto="no" failed="no" names=""
id="0.0.1547" type="dasd-eckd" on="no" exists="yes" pers="no" auto="no" failed="no" names=""
id="0.0.1548" type="dasd-eckd" on="no" exists="yes" pers="no" auto="no" failed="no" names=""
id="0.0.1549" type="dasd-eckd" on="no" exists="yes" pers="no" auto="no" failed="no" names=""
id="0.0.154a" type="dasd-eckd" on="no" exists="yes" pers="no" auto="no" failed="no" names=""
id="0.0.154b" type="dasd-eckd" on="no" exists="yes" pers="no" auto="no" failed="no" names=""
id="0.0.154c" type="dasd-eckd" on="no" exists="yes" pers="no" auto="no" failed="no" names=""
id="0.0.154d" type="dasd-eckd" on="yes" exists="yes" pers="yes" auto="no" failed="no" names="dasda"
id="0.0.154e" type="dasd-eckd" on="yes" exists="yes" pers="no" auto="yes" failed="no" names="dasdb"
id="0.0.154f" type="dasd-eckd" on="no" exists="yes" pers="no" auto="no" failed="no" names=""
id="0.0.1550" type="dasd-eckd" on="no" exists="yes" pers="no" auto="no" failed="no" names=""
id="0.0.1551" type="dasd-eckd" on="no" exists="yes" pers="no" auto="no" failed="no" names=""
id="0.0.1552" type="dasd-eckd" on="no" exists="yes" pers="no" auto="no" failed="no" names=""
id="0.0.1553" type="dasd-eckd" on="no" exists="yes" pers="no" auto="no" failed="no" names=""
id="0.0.1554" type="dasd-eckd" on="no" exists="yes" pers="no" auto="no" failed="no" names=""
id="0.0.1555" type="dasd-eckd" on="no" exists="yes" pers="no" auto="no" failed="no" names=""
id="0.0.15e0" type="dasd-eckd" on="no" exists="yes" pers="no" auto="no" failed="no" names=""
id="0.0.15e1" type="dasd-eckd" on="no" exists="yes" pers="no" auto="no" failed="no" names=""
id="0.0.15e2" type="dasd-eckd" on="no" exists="yes" pers="no" auto="no" failed="no" names=""
id="0.0.15e3" type="dasd-eckd" on="no" exists="yes" pers="no" auto="no" failed="no" names=""
id="0.0.15e4" type="dasd-eckd" on="no" exists="yes" pers="no" auto="no" failed="no" names=""
id="0.0.15e5" type="dasd-eckd" on="no" exists="yes" pers="no" auto="no" failed="no" names=""
id="0.0.15e6" type="dasd-eckd" on="no" exists="yes" pers="no" auto="no" failed="no" names=""
id="0.0.15e7" type="dasd-eckd" on="no" exists="yes" pers="no" auto="no" failed="no" names=""
id="0.0.15e8" type="dasd-eckd" on="no" exists="yes" pers="no" auto="no" failed="no" names=""
id="0.0.15e9" type="dasd-eckd" on="no" exists="yes" pers="no" auto="no" failed="no" names=""
id="0.0.15ea" type="dasd-eckd" on="no" exists="yes" pers="no" auto="no" failed="no" names=""
id="0.0.15eb" type="dasd-eckd" on="no" exists="yes" pers="no" auto="no" failed="no" names=""
id="0.0.15ec" type="dasd-eckd" on="no" exists="yes" pers="no" auto="no" failed="no" names=""
id="0.0.15ed" type="dasd-eckd" on="no" exists="yes" pers="no" auto="no" failed="no" names=""
id="0.0.15ee" type="dasd-eckd" on="no" exists="yes" pers="no" auto="no" failed="no" names=""
id="0.0.15ef" type="dasd-eckd" on="no" exists="yes" pers="no" auto="no" failed="no" names=""
id="0.0.15f0" type="dasd-eckd" on="no" exists="yes" pers="no" auto="no" failed="no" names=""
id="0.0.15f1" type="dasd-eckd" on="no" exists="yes" pers="no" auto="no" failed="no" names=""
id="0.0.15f2" type="dasd-eckd" on="no" exists="yes" pers="no" auto="no" failed="no" names=""
id="0.0.15f3" type="dasd-eckd" on="no" exists="yes" pers="no" auto="no" failed="no" names=""
id="0.0.15f4" type="dasd-eckd" on="no" exists="yes" pers="no" auto="no" failed="no" names=""
id="0.0.15f5" type="dasd-eckd" on="no" exists="yes" pers="no" auto="no" failed="no" names=""
id="0.0.15f6" type="dasd-eckd" on="no" exists="yes" pers="no" auto="no" failed="no" names=""
id="0.0.15f7" type="dasd-eckd" on="no" exists="yes" pers="no" auto="no" failed="no" names=""
id="0.0.15f8" type="dasd-eckd" on="no" exists="yes" pers="no" auto="no" failed="no" names=""
id="0.0.15f9" type="dasd-eckd" on="no" exists="yes" pers="no" auto="no" failed="no" names=""
id="0.0.15fa" type="dasd-eckd" on="no" exists="yes" pers="no" auto="no" failed="no" names=""
id="0.0.15fb" type="dasd-eckd" on="no" exists="yes" pers="no" auto="no" failed="no" names=""
id="0.0.15fc" type="dasd-eckd" on="no" exists="yes" pers="no" auto="no" failed="no" names=""
id="0.0.15fd" type="dasd-eckd" on="no" exists="yes" pers="no" auto="no" failed="no" names=""
id="0.0.15fe" type="dasd-eckd" on="no" exists="yes" pers="no" auto="no" failed="no" names=""
id="0.0.15ff" type="dasd-eckd" on="no" exists="yes" pers="no" auto="no" failed="no" names=""
id="0.0.2500" type="dasd-eckd" on="no" exists="yes" pers="no" auto="no" failed="no" names=""
id="0.0.2501" type="dasd-eckd" on="no" exists="yes" pers="no" auto="no" failed="no" names=""
id="0.0.2502" type="dasd-eckd" on="no" exists="yes" pers="no" auto="no" failed="no" names=""
id="0.0.2503" type="dasd-eckd" on="no" exists="yes" pers="no" auto="no" failed="no" names=""
id="0.0.2504" type="dasd-eckd" on="no" exists="yes" pers="no" auto="no" failed="no" names=""
id="0.0.2505" type="dasd-eckd" on="no" exists="yes" pers="no" auto="no" failed="no" names=""
id="0.0.2506" type="dasd-eckd" on="no" exists="yes" pers="no" auto="no" failed="no" names=""
id="0.0.2507" type="dasd-eckd" on="no" exists="yes" pers="no" auto="no" failed="no" names=""
id="0.0.2508" type="dasd-eckd" on="no" exists="yes" pers="no" auto="no" failed="no" names=""
id="0.0.2509" type="dasd-eckd" on="no" exists="yes" pers="no" auto="no" failed="no" names=""
id="0.0.250a" type="dasd-eckd" on="no" exists="yes" pers="no" auto="no" failed="no" names=""
id="0.0.250b" type="dasd-eckd" on="no" exists="yes" pers="no" auto="no" failed="no" names=""
id="0.0.250c" type="dasd-eckd" on="no" exists="yes" pers="no" auto="no" failed="no" names=""
id="0.0.250d" type="dasd-eckd" on="no" exists="yes" pers="no" auto="no" failed="no" names=""
id="0.0.250e" type="dasd-eckd" on="no" exists="yes" pers="no" auto="no" failed="no" names=""
id="0.0.250f" type="dasd-eckd" on="no" exists="yes" pers="no" auto="no" failed="no" names=""
id="0.0.2510" type="dasd-eckd" on="no" exists="yes" pers="no" auto="no" failed="no" names=""
id="0.0.2511" type="dasd-eckd" on="no" exists="yes" pers="no" auto="no" failed="no" names=""
id="0.0.2512" type="dasd-eckd" on="no" exists="yes" pers="no" auto="no" failed="no" names=""
id="0.0.2513" type="dasd-eckd" on="no" exists="yes" pers="no" auto="no" failed="no" names=""
id="0.0.2514" type="dasd-eckd" on="no" exists="yes" pers="no" auto="no" failed="no" names=""
id="0.0.2515" type="dasd-eckd" on="no" exists="yes" pers="no" auto="no" failed="no" names=""
id="0.0.2516" type="dasd-eckd" on="no" exists="yes" pers="no" auto="no" failed="no" names=""
id="0.0.2517" type="dasd-eckd" on="no" exists="yes" pers="no" auto="no" failed="no" names=""
id="0.0.2518" type="dasd-eckd" on="no" exists="yes" pers="no" auto="no" failed="no" names=""
id="0.0.2519" type="dasd-eckd" on="no" exists="yes" pers="no" auto="no" failed="no" names=""
id="0.0.251a" type="dasd-eckd" on="no" exists="yes" pers="no" auto="no" failed="no" names=""
id="0.0.251b" type="dasd-eckd" on="no" exists="yes" pers="no" auto="no" failed="no" names=""
id="0.0.251c" type="dasd-eckd" on="no" exists="yes" pers="no" auto="no" failed="no" names=""
id="0.0.251d" type="dasd-eckd" on="no" exists="yes" pers="no" auto="no" failed="no" names=""
id="0.0.251e" type="dasd-eckd" on="no" exists="yes" pers="no" auto="no" failed="no" names=""
id="0.0.251f" type="dasd-eckd" on="no" exists="yes" pers="no" auto="no" failed="no" names=""
id="0.0.2520" type="dasd-eckd" on="no" exists="yes" pers="no" auto="no" failed="no" names=""
id="0.0.2521" type="dasd-eckd" on="no" exists="yes" pers="no" auto="no" failed="no" names=""
id="0.0.2522" type="dasd-eckd" on="no" exists="yes" pers="no" auto="no" failed="no" names=""
id="0.0.2523" type="dasd-eckd" on="no" exists="yes" pers="no" auto="no" failed="no" names=""
id="0.0.2524" type="dasd-eckd" on="no" exists="yes" pers="no" auto="no" failed="no" names=""
id="0.0.2525" type="dasd-eckd" on="no" exists="yes" pers="no" auto="no" failed="no" names=""
id="0.0.2526" type="dasd-eckd" on="no" exists="yes" pers="no" auto="no" failed="no" names=""
id="0.0.2527" type="dasd-eckd" on="no" exists="yes" pers="no" auto="no" failed="no" names=""
id="0.0.2528" type="dasd-eckd" on="no" exists="yes" pers="no" auto="no" failed="no" names=""
id="0.0.2529" type="dasd-eckd" on="no" exists="yes" pers="no" auto="no" failed="no" names=""
id="0.0.252a" type="dasd-eckd" on="no" exists="yes" pers="no" auto="no" failed="no" names=""
id="0.0.252b" type="dasd-eckd" on="no" exists="yes" pers="no" auto="no" failed="no" names=""
id="0.0.252c" type="dasd-eckd" on="no" exists="yes" pers="no" auto="no" failed="no" names=""
id="0.0.252d" type="dasd-eckd" on="no" exists="yes" pers="no" auto="no" failed="no" names=""
id="0.0.252e" type="dasd-eckd" on="no" exists="yes" pers="no" auto="no" failed="no" names=""
id="0.0.252f" type="dasd-eckd" on="no" exists="yes" pers="no" auto="no" failed="no" names=""
id="0.0.2530" type="dasd-eckd" on="no" exists="yes" pers="no" auto="no" failed="no" names=""
id="0.0.2531" type="dasd-eckd" on="no" exists="yes" pers="no" auto="no" failed="no" names=""
id="0.0.2532" type="dasd-eckd" on="no" exists="yes" pers="no" auto="no" failed="no" names=""
id="0.0.2533" type="dasd-eckd" on="no" exists="yes" pers="no" auto="no" failed="no" names=""
id="0.0.2534" type="dasd-eckd" on="no" exists="yes" pers="no" auto="no" failed="no" names=""
id="0.0.2535" type="dasd-eckd" on="no" exists="yes" pers="no" auto="no" failed="no" names=""
id="0.0.2536" type="dasd-eckd" on="no" exists="yes" pers="no" auto="no" failed="no" names=""
id="0.0.2537" type="dasd-eckd" on="no" exists="yes" pers="no" auto="no" failed="no" names=""
id="0.0.2538" type="dasd-eckd" on="no" exists="yes" pers="no" auto="no" failed="no" names=""
id="0.0.2539" type="dasd-eckd" on="no" exists="yes" pers="no" auto="no" failed="no" names=""
id="0.0.253a" type="dasd-eckd" on="no" exists="yes" pers="no" auto="no" failed="no" names=""
id="0.0.253b" type="dasd-eckd" on="no" exists="yes" pers="no" auto="no" failed="no" names=""
id="0.0.253c" type="dasd-eckd" on="no" exists="yes" pers="no" auto="no" failed="no" names=""
id="0.0.253d" type="dasd-eckd" on="no" exists="yes" pers="no" auto="no" failed="no" names=""
id="0.0.253e" type="dasd-eckd" on="no" exists="yes" pers="no" auto="no" failed="no" names=""
id="0.0.253f" type="dasd-eckd" on="no" exists="yes" pers="no" auto="no" failed="no" names=""
id="0.0.2540" type="dasd-eckd" on="no" exists="yes" pers="no" auto="no" failed="no" names=""
id="0.0.2541" type="dasd-eckd" on="no" exists="yes" pers="no" auto="no" failed="no" names=""
id="0.0.2542" type="dasd-eckd" on="no" exists="yes" pers="no" auto="no" failed="no" names=""
id="0.0.2543" type="dasd-eckd" on="no" exists="yes" pers="no" auto="no" failed="no" names=""
id="0.0.2544" type="dasd-eckd" on="no" exists="yes" pers="no" auto="no" failed="no" names=""
id="0.0.2545" type="dasd-eckd" on="no" exists="yes" pers="no" auto="no" failed="no" names=""
id="0.0.2546" type="dasd-eckd" on="no" exists="yes" pers="no" auto="no" failed="no" names=""
id="0.0.2547" type="dasd-eckd" on="no" exists="yes" pers="no" auto="no" failed="no" names=""
id="0.0.2548" type="dasd-eckd" on="no" exists="yes" pers="no" auto="no" failed="no" names=""
id="0.0.2549" type="dasd-eckd" on="no" exists="yes" pers="no" auto="no" failed="no" names=""
id="0.0.254a" type="dasd-eckd" on="no" exists="yes" pers="no" auto="no" failed="no" names=""
id="0.0.254b" type="dasd-eckd" on="no" exists="yes" pers="no" auto="no" failed="no" names=""
id="0.0.254c" type="dasd-eckd" on="no" exists="yes" pers="no" auto="no" failed="no" names=""
id="0.0.254d" type="dasd-eckd" on="no" exists="yes" pers="no" auto="no" failed="no" names=""
id="0.0.254e" type="dasd-eckd" on="no" exists="yes" pers="no" auto="no" failed="no" names=""
id="0.0.254f" type="dasd-eckd" on="no" exists="yes" pers="no" auto="no" failed="no" names=""
id="0.0.2550" type="dasd-eckd" on="no" exists="yes" pers="no" auto="no" failed="no" names=""
id="0.0.2551" type="dasd-eckd" on="no" exists="yes" pers="no" auto="no" failed="no" names=""
id="0.0.2552" type="dasd-eckd" on="no" exists="yes" pers="no" auto="no" failed="no" names=""
id="0.0.2553" type="dasd-eckd" on="no" exists="yes" pers="no" auto="no" failed="no" names=""
id="0.0.2554" type="dasd-eckd" on="no" exists="yes" pers="no" auto="no" failed="no" names=""
id="0.0.2555" type="dasd-eckd" on="no" exists="yes" pers="no" auto="no" failed="no" names=""
id="0.0.e000" type="zfcp-host" on="yes" exists="yes" pers="yes" auto="no" failed="no" names=""
id="0.0.e001" type="zfcp-host" on="no" exists="yes" pers="no" auto="no" failed="no" names=""
id="0.0.e002" type="zfcp-host" on="no" exists="yes" pers="no" auto="no" failed="no" names=""
id="0.0.e003" type="zfcp-host" on="no" exists="yes" pers="no" auto="no" failed="no" names=""
id="0.0.e004" type="zfcp-host" on="no" exists="yes" pers="no" auto="no" failed="no" names=""
id="0.0.e005" type="zfcp-host" on="no" exists="yes" pers="no" auto="no" failed="no" names=""
id="0.0.e006" type="zfcp-host" on="no" exists="yes" pers="no" auto="no" failed="no" names=""
id="0.0.e007" type="zfcp-host" on="no" exists="yes" pers="no" auto="no" failed="no" names=""
id="0.0.e008" type="zfcp-host" on="no" exists="yes" pers="no" auto="no" failed="no" names=""
id="0.0.e009" type="zfcp-host" on="no" exists="yes" pers="no" auto="no" failed="no" names=""
id="0.0.e00a" type="zfcp-host" on="no" exists="yes" pers="no" auto="no" failed="no" names=""
id="0.0.e00b" type="zfcp-host" on="no" exists="yes" pers="no" auto="no" failed="no" names=""
id="0.0.e00c" type="zfcp-host" on="no" exists="yes" pers="no" auto="no" failed="no" names=""
id="0.0.e00d" type="zfcp-host" on="no" exists="yes" pers="no" auto="no" failed="no" names=""
id="0.0.e00e" type="zfcp-host" on="no" exists="yes" pers="no" auto="no" failed="no" names=""
id="0.0.e00f" type="zfcp-host" on="no" exists="yes" pers="no" auto="no" failed="no" names=""
id="0.0.e100" type="zfcp-host" on="no" exists="yes" pers="no" auto="no" failed="no" names=""
id="0.0.e101" type="zfcp-host" on="no" exists="yes" pers="no" auto="no" failed="no" names=""
id="0.0.e102" type="zfcp-host" on="no" exists="yes" pers="no" auto="no" failed="no" names=""
id="0.0.e103" type="zfcp-host" on="no" exists="yes" pers="no" auto="no" failed="no" names=""
id="0.0.e104" type="zfcp-host" on="no" exists="yes" pers="no" auto="no" failed="no" names=""
id="0.0.e105" type="zfcp-host" on="no" exists="yes" pers="no" auto="no" failed="no" names=""
id="0.0.e106" type="zfcp-host" on="no" exists="yes" pers="no" auto="no" failed="no" names=""
id="0.0.e107" type="zfcp-host" on="no" exists="yes" pers="no" auto="no" failed="no" names=""
id="0.0.e108" type="zfcp-host" on="no" exists="yes" pers="no" auto="no" failed="no" names=""
id="0.0.e109" type="zfcp-host" on="no" exists="yes" pers="no" auto="no" failed="no" names=""
id="0.0.e10a" type="zfcp-host" on="no" exists="yes" pers="no" auto="no" failed="no" names=""
id="0.0.e10b" type="zfcp-host" on="no" exists="yes" pers="no" auto="no" failed="no" names=""
id="0.0.e10c" type="zfcp-host" on="no" exists="yes" pers="no" auto="no" failed="no" names=""
id="0.0.e10d" type="zfcp-host" on="no" exists="yes" pers="no" auto="no" failed="no" names=""
id="0.0.e10e" type="zfcp-host" on="no" exists="yes" pers="no" auto="no" failed="no" names=""
id="0.0.e10f" type="zfcp-host" on="no" exists="yes" pers="no" auto="no" failed="no" names=""
id="0.0.e000:0x50050763060b16b6:0x4024400400000000" type="zfcp-lun" on="yes" exists="yes" pers="no" auto="no" failed="no" names="sda sg0"
id="0.0.e000:0x50050763060b16b6:0x4024400500000000" type="zfcp-lun" on="yes" exists="yes" pers="no" auto="no" failed="no" names="sdb sg1"
id="0.0.e000:0x50050763061b16b6:0x4024400400000000" type="zfcp-lun" on="yes" exists="yes" pers="no" auto="no" failed="no" names="sdc sg2"
id="0.0.e000:0x50050763061b16b6:0x4024400500000000" type="zfcp-lun" on="yes" exists="yes" pers="no" auto="no" failed="no" names="sdd sg3"
id="0.0.c000:0.0.c001:0.0.c002" type="qeth" on="no" exists="yes" pers="no" auto="no" failed="no" names=""
id="0.0.c003:0.0.c004:0.0.c005" type="qeth" on="yes" exists="yes" pers="yes" auto="no" failed="no" names="encc003"
id="0.0.c006:0.0.c007:0.0.c008" type="qeth" on="no" exists="yes" pers="no" auto="no" failed="no" names=""
id="0.0.c009:0.0.c00a:0.0.c00b" type="qeth" on="no" exists="yes" pers="no" auto="no" failed="no" names=""
id="0.0.c00c:0.0.c00d:0.0.c00e" type="qeth" on="no" exists="yes" pers="no" auto="no" failed="no" names=""
id="0.0.c00f:0.0.c010:0.0.c011" type="qeth" on="no" exists="yes" pers="no" auto="no" failed="no" names=""
id="0.0.c012:0.0.c013:0.0.c014" type="qeth" on="no" exists="yes" pers="no" auto="no" failed="no" names=""
id="0.0.c015:0.0.c016:0.0.c017" type="qeth" on="no" exists="yes" pers="no" auto="no" failed="no" names=""
id="0.0.c018:0.0.c019:0.0.c01a" type="qeth" on="no" exists="yes" pers="no" auto="no" failed="no" names=""
id="0.0.c01b:0.0.c01c:0.0.c01d" type="qeth" on="no" exists="yes" pers="no" auto="no" failed="no" names=""
id="0.0.c01e:0.0.c01f:0.0.c020" type="qeth" on="no" exists="yes" pers="no" auto="no" failed="no" names=""
id="0.0.c021:0.0.c022:0.0.c023" type="qeth" on="no" exists="yes" pers="no" auto="no" failed="no" names=""
id="0.0.c024:0.0.c025:0.0.c026" type="qeth" on="no" exists="yes" pers="no" auto="no" failed="no" names=""
id="0.0.c027:0.0.c028:0.0.c029" type="qeth" on="no" exists="yes" pers="no" auto="no" failed="no" names=""
id="0.0.c02a:0.0.c02b:0.0.c02c" type="qeth" on="no" exists="yes" pers="no" auto="no" failed="no" names=""
id="0.0.c02d:0.0.c02e:0.0.c02f" type="qeth" on="no" exists="yes" pers="no" auto="no" failed="no" names=""
id="0.0.c030:0.0.c031:0.0.c032" type="qeth" on="no" exists="yes" pers="no" auto="no" failed="no" names=""
id="0.0.c033:0.0.c034:0.0.c035" type="qeth" on="no" exists="yes" pers="no" auto="no" failed="no" names=""
id="0.0.c036:0.0.c037:0.0.c038" type="qeth" on="no" exists="yes" pers="no" auto="no" failed="no" names=""
id="0.0.c039:0.0.c03a:0.0.c03b" type="qeth" on="no" exists="yes" pers="no" auto="no" failed="no" names=""
id="0.0.c03c:0.0.c03d:0.0.c03e" type="qeth" on="no" exists="yes" pers="no" auto="no" failed="no" names=""
id="0.0.c03f:0.0.c040:0.0.c041" type="qeth" on="no" exists="yes" pers="no" auto="no" failed="no" names=""
id="0.0.c042:0.0.c043:0.0.c044" type="qeth" on="no" exists="yes" pers="no" auto="no" failed="no" names=""
id="0.0.c045:0.0.c046:0.0.c047" type="qeth" on="no" exists="yes" pers="no" auto="no" failed="no" names=""
id="0.0.c048:0.0.c049:0.0.c04a" type="qeth" on="no" exists="yes" pers="no" auto="no" failed="no" names=""
id="0.0.c04b:0.0.c04c:0.0.c04d" type="qeth" on="no" exists="yes" pers="no" auto="no" failed="no" names=""
id="0.0.c04e:0.0.c04f:0.0.c050" type="qeth" on="no" exists="yes" pers="no" auto="no" failed="no" names=""
id="0.0.c051:0.0.c052:0.0.c053" type="qeth" on="no" exists="yes" pers="no" auto="no" failed="no" names=""
id="0.0.c054:0.0.c055:0.0.c056" type="qeth" on="no" exists="yes" pers="no" auto="no" failed="no" names=""
id="0.0.c057:0.0.c058:0.0.c059" type="qeth" on="no" exists="yes" pers="no" auto="no" failed="no" names=""
id="0.0.c05a:0.0.c05b:0.0.c05c" type="qeth" on="no" exists="yes" pers="no" auto="no" failed="no" names=""
id="0.0.c05d:0.0.c05e:0.0.c05f" type="qeth" on="no" exists="yes" pers="no" auto="no" failed="no" names=""
id="0.0.c060:0.0.c061:0.0.c062" type="qeth" on="no" exists="yes" pers="no" auto="no" failed="no" names=""
id="0.0.c063:0.0.c064:0.0.c065" type="qeth" on="no" exists="yes" pers="no" auto="no" failed="no" names=""
id="0.0.c066:0.0.c067:0.0.c068" type="qeth" on="no" exists="yes" pers="no" auto="no" failed="no" names=""
id="0.0.c069:0.0.c06a:0.0.c06b" type="qeth" on="no" exists="yes" pers="no" auto="no" failed="no" names=""
id="0.0.c06c:0.0.c06d:0.0.c06e" type="qeth" on="no" exists="yes" pers="no" auto="no" failed="no" names=""
id="0.0.c06f:0.0.c070:0.0.c071" type="qeth" on="no" exists="yes" pers="no" auto="no" failed="no" names=""
id="0.0.c072:0.0.c073:0.0.c074" type="qeth" on="no" exists="yes" pers="no" auto="no" failed="no" names=""
id="0.0.c075:0.0.c076:0.0.c077" type="qeth" on="no" exists="yes" pers="no" auto="no" failed="no" names=""
id="0.0.c078:0.0.c079:0.0.c07a" type="qeth" on="no" exists="yes" pers="no" auto="no" failed="no" names=""
id="0.0.c07b:0.0.c07c:0.0.c07d" type="qeth" on="no" exists="yes" pers="no" auto="no" failed="no" names=""
id="0.0.d000:0.0.d001:0.0.d002" type="qeth" on="no" exists="yes" pers="no" auto="no" failed="no" names=""
id="0.0.d003:0.0.d004:0.0.d005" type="qeth" on="no" exists="yes" pers="no" auto="no" failed="no" names=""
id="0.0.d006:0.0.d007:0.0.d008" type="qeth" on="no" exists="yes" pers="no" auto="no" failed="no" names=""
id="0.0.d009:0.0.d00a:0.0.d00b" type="qeth" on="no" exists="yes" pers="no" auto="no" failed="no" names=""
id="0.0.d00c:0.0.d00d:0.0.d00e" type="qeth" on="no" exists="yes" pers="no" auto="no" failed="no" names=""
id="0.0.d00f:0.0.d010:0.0.d011" type="qeth" on="no" exists="yes" pers="no" auto="no" failed="no" names=""
id="0.0.d012:0.0.d013:0.0.d014" type="qeth" on="no" exists="yes" pers="no" auto="no" failed="no" names=""
id="0.0.d015:0.0.d016:0.0.d017" type="qeth" on="no" exists="yes" pers="no" auto="no" failed="no" names=""
id="0.0.d018:0.0.d019:0.0.d01a" type="qeth" on="no" exists="yes" pers="no" auto="no" failed="no" names=""
id="0.0.d01b:0.0.d01c:0.0.d01d" type="qeth" on="no" exists="yes" pers="no" auto="no" failed="no" names=""
id="0.0.d01e:0.0.d01f:0.0.d020" type="qeth" on="no" exists="yes" pers="no" auto="no" failed="no" names=""
id="0.0.d021:0.0.d022:0.0.d023" type="qeth" on="no" exists="yes" pers="no" auto="no" failed="no" names=""
id="0.0.d024:0.0.d025:0.0.d026" type="qeth" on="no" exists="yes" pers="no" auto="no" failed="no" names=""
id="0.0.d027:0.0.d028:0.0.d029" type="qeth" on="no" exists="yes" pers="no" auto="no" failed="no" names=""
id="0.0.d02a:0.0.d02b:0.0.d02c" type="qeth" on="no" exists="yes" pers="no" auto="no" failed="no" names=""
id="0.0.d02d:0.0.d02e:0.0.d02f" type="qeth" on="no" exists="yes" pers="no" auto="no" failed="no" names=""
id="0.0.d030:0.0.d031:0.0.d032" type="qeth" on="no" exists="yes" pers="no" auto="no" failed="no" names=""
id="0.0.d033:0.0.d034:0.0.d035" type="qeth" on="no" exists="yes" pers="no" auto="no" failed="no" names=""
id="0.0.d036:0.0.d037:0.0.d038" type="qeth" on="no" exists="yes" pers="no" auto="no" failed="no" names=""
id="0.0.d039:0.0.d03a:0.0.d03b" type="qeth" on="no" exists="yes" pers="no" auto="no" failed="no" names=""
id="0.0.d03c:0.0.d03d:0.0.d03e" type="qeth" on="no" exists="yes" pers="no" auto="no" failed="no" names=""
id="0.0.d03f:0.0.d040:0.0.d041" type="qeth" on="no" exists="yes" pers="no" auto="no" failed="no" names=""
id="0.0.d042:0.0.d043:0.0.d044" type="qeth" on="no" exists="yes" pers="no" auto="no" failed="no" names=""
id="0.0.d045:0.0.d046:0.0.d047" type="qeth" on="no" exists="yes" pers="no" auto="no" failed="no" names=""
id="0.0.d100:0.0.d101:0.0.d102" type="qeth" on="no" exists="yes" pers="no" auto="no" failed="no" names=""
id="0.0.d103:0.0.d104:0.0.d105" type="qeth" on="no" exists="yes" pers="no" auto="no" failed="no" names=""
id="0.0.d106:0.0.d107:0.0.d108" type="qeth" on="no" exists="yes" pers="no" auto="no" failed="no" names=""
id="0.0.d109:0.0.d10a:0.0.d10b" type="qeth" on="no" exists="yes" pers="no" auto="no" failed="no" names=""
id="0.0.d10c:0.0.d10d:0.0.d10e" type="qeth" on="no" exists="yes" pers="no" auto="no" failed="no" names=""
id="0.0.d10f:0.0.d110:0.0.d111" type="qeth" on="no" exists="yes" pers="no" auto="no" failed="no" names=""
id="0.0.d112:0.0.d113:0.0.d114" type="qeth" on="no" exists="yes" pers="no" auto="no" failed="no" names=""
id="0.0.d115:0.0.d116:0.0.d117" type="qeth" on="no" exists="yes" pers="no" auto="no" failed="no" names=""
id="0.0.d118:0.0.d119:0.0.d11a" type="qeth" on="no" exists="yes" pers="no" auto="no" failed="no" names=""
id="0.0.d11b:0.0.d11c:0.0.d11d" type="qeth" on="no" exists="yes" pers="no" auto="no" failed="no" names=""
id="0.0.d11e:0.0.d11f:0.0.d120" type="qeth" on="no" exists="yes" pers="no" auto="no" failed="no" names=""
id="0.0.d121:0.0.d122:0.0.d123" type="qeth" on="no" exists="yes" pers="no" auto="no" failed="no" names=""
id="0.0.d124:0.0.d125:0.0.d126" type="qeth" on="no" exists="yes" pers="no" auto="no" failed="no" names=""
id="0.0.d127:0.0.d128:0.0.d129" type="qeth" on="no" exists="yes" pers="no" auto="no" failed="no" names=""
id="0.0.d12a:0.0.d12b:0.0.d12c" type="qeth" on="no" exists="yes" pers="no" auto="no" failed="no" names=""
id="0.0.d12d:0.0.d12e:0.0.d12f" type="qeth" on="no" exists="yes" pers="no" auto="no" failed="no" names=""
id="0.0.d130:0.0.d131:0.0.d132" type="qeth" on="no" exists="yes" pers="no" auto="no" failed="no" names=""
id="0.0.d133:0.0.d134:0.0.d135" type="qeth" on="no" exists="yes" pers="no" auto="no" failed="no" names=""
id="0.0.d136:0.0.d137:0.0.d138" type="qeth" on="no" exists="yes" pers="no" auto="no" failed="no" names=""
id="0.0.d139:0.0.d13a:0.0.d13b" type="qeth" on="no" exists="yes" pers="no" auto="no" failed="no" names=""
id="0.0.d13c:0.0.d13d:0.0.d13e" type="qeth" on="no" exists="yes" pers="no" auto="no" failed="no" names=""
id="0.0.d13f:0.0.d140:0.0.d141" type="qeth" on="no" exists="yes" pers="no" auto="no" failed="no" names=""
id="0.0.d142:0.0.d143:0.0.d144" type="qeth" on="no" exists="yes" pers="no" auto="no" failed="no" names=""
id="0.0.d145:0.0.d146:0.0.d147" type="qeth" on="no" exists="yes" pers="no" auto="no" failed="no" names=""
id="0.0.a110:0.0.a111" type="ctc" on="no" exists="yes" pers="no" auto="no" failed="no" names=""
id="0.0.a112:0.0.a113" type="ctc" on="no" exists="yes" pers="no" auto="no" failed="no" names=""
id="0.0.a114:0.0.a115" type="ctc" on="no" exists="yes" pers="no" auto="no" failed="no" names=""
id="0.0.a116:0.0.a117" type="ctc" on="no" exists="yes" pers="no" auto="no" failed="no" names=""
id="0.0.a118:0.0.a119" type="ctc" on="no" exists="yes" pers="no" auto="no" failed="no" names=""
id="0.0.a11a:0.0.a11b" type="ctc" on="no" exists="yes" pers="no" auto="no" failed="no" names=""
id="0.0.a11c:0.0.a11d" type="ctc" on="no" exists="yes" pers="no" auto="no" failed="no" names=""
id="0.0.a11e:0.0.a11f" type="ctc" on="no" exists="yes" pers="no" auto="no" failed="no" names=""
id="0.0.a120:0.0.a121" type="ctc" on="no" exists="yes" pers="no" auto="no" failed="no" names=""
id="0.0.a122:0.0.a123" type="ctc" on="no" exists="yes" pers="no" auto="no" failed="no" names=""
id="0.0.a124:0.0.a125" type="ctc" on="no" exists="yes" pers="no" auto="no" failed="no" names=""
id="0.0.a126:0.0.a127" type="ctc" on="no" exists="yes" pers="no" auto="no" failed="no" names=""
id="0.0.a128:0.0.a129" type="ctc" on="no" exists="yes" pers="no" auto="no" failed="no" names=""
id="0.0.a12a:0.0.a12b" type="ctc" on="no" exists="yes" pers="no" auto="no" failed="no" names=""
id="0.0.a12c:0.0.a12d" type="ctc" on="no" exists="yes" pers="no" auto="no" failed="no" names=""
id="0.0.a12e:0.0.a12f" type="ctc" on="no" exists="yes" pers="no" auto="no" failed="no" names=""
id="0.0.a130:0.0.a131" type="ctc" on="no" exists="yes" pers="no" auto="no" failed="no" names=""
id="0.0.a132:0.0.a133" type="ctc" on="no" exists="yes" pers="no" auto="no" failed="no" names=""
id="0.0.a134:0.0.a135" type="ctc" on="no" exists="yes" pers="no" auto="no" failed="no" names=""
id="0.0.a136:0.0.a137" type="ctc" on="no" exists="yes" pers="no" auto="no" failed="no" names=""
id="0.0.a138:0.0.a139" type="ctc" on="no" exists="yes" pers="no" auto="no" failed="no" names=""
id="0.0.a13a:0.0.a13b" type="ctc" on="no" exists="yes" pers="no" auto="no" failed="no" names=""
id="0.0.a13c:0.0.a13d" type="ctc" on="no" exists="yes" pers="no" auto="no" failed="no" names=""
id="0.0.a13e:0.0.a13f" type="ctc" on="no" exists="yes" pers="no" auto="no" failed="no" names=""
id="0.0.a140:0.0.a141" type="ctc" on="no" exists="yes" pers="no" auto="no" failed="no" names=""
id="0.0.a142:0.0.a143" type="ctc" on="no" exists="yes" pers="no" auto="no" failed="no" names=""
id="0.0.a144:0.0.a145" type="ctc" on="no" exists="yes" pers="no" auto="no" failed="no" names=""
id="0.0.a146:0.0.a147" type="ctc" on="no" exists="yes" pers="no" auto="no" failed="no" names=""
id="0.0.a148:0.0.a149" type="ctc" on="no" exists="yes" pers="no" auto="no" failed="no" names=""
id="0.0.a14a:0.0.a14b" type="ctc" on="no" exists="yes" pers="no" auto="no" failed="no" names=""
id="0.0.a14c:0.0.a14d" type="ctc" on="no" exists="yes" pers="no" auto="no" failed="no" names=""
id="0.0.a14e:0.0.a14f" type="ctc" on="no" exists="yes" pers="no" auto="no" failed="no" names=""
id="0.0.a150:0.0.a151" type="ctc" on="no" exists="yes" pers="no" auto="no" failed="no" names=""
id="0.0.a152:0.0.a153" type="ctc" on="no" exists="yes" pers="no" auto="no" failed="no" names=""
id="0.0.a154:0.0.a155" type="ctc" on="no" exists="yes" pers="no" auto="no" failed="no" names=""
id="0.0.a156:0.0.a157" type="ctc" on="no" exists="yes" pers="no" auto="no" failed="no" names=""
id="0.0.a158:0.0.a159" type="ctc" on="no" exists="yes" pers="no" auto="no" failed="no" names=""
id="0.0.a15a:0.0.a15b" type="ctc" on="no" exists="yes" pers="no" auto="no" failed="no" names=""
id="0.0.a15c:0.0.a15d" type="ctc" on="no" exists="yes" pers="no" auto="no" failed="no" names=""
id="0.0.a15e:0.0.a15f" type="ctc" on="no" exists="yes" pers="no" auto="no" failed="no" names=""
id="0.0.a160:0.0.a161" type="ctc" on="no" exists="yes" pers="no" auto="no" failed="no" names=""
id="0.0.a162:0.0.a163" type="ctc" on="no" exists="yes" pers="no" auto="no" failed="no" names=""
id="0.0.a164:0.0.a165" type="ctc" on="no" exists="yes" pers="no" auto="no" failed="no" names=""
id="0.0.a166:0.0.a167" type="ctc" on="no" exists="yes" pers="no" auto="no" failed="no" names=""
id="0.0.a168:0.0.a169" type="ctc" on="no" exists="yes" pers="no" auto="no" failed="no" names=""
id="0.0.a16a:0.0.a16b" type="ctc" on="no" exists="yes" pers="no" auto="no" failed="no" names=""
id="0.0.a16c:0.0.a16d" type="ctc" on="no" exists="yes" pers="no" auto="no" failed="no" names=""
id="0.0.a16e:0.0.a16f" type="ctc" on="no" exists="yes" pers="no" auto="no" failed="no" names=""
id="0.0.a180:0.0.a181" type="ctc" on="no" exists="yes" pers="no" auto="no" failed="no" names=""
id="0.0.a182:0.0.a183" type="ctc" on="no" exists="yes" pers="no" auto="no" failed="no" names=""
id="0.0.a184:0.0.a185" type="ctc" on="no" exists="yes" pers="no" auto="no" failed="no" names=""
id="0.0.a186:0.0.a187" type="ctc" on="no" exists="yes" pers="no" auto="no" failed="no" names=""
id="0.0.a188:0.0.a189" type="ctc" on="no" exists="yes" pers="no" auto="no" failed="no" names=""
id="0.0.a18a:0.0.a18b" type="ctc" on="no" exists="yes" pers="no" auto="no" failed="no" names=""
id="0.0.a18c:0.0.a18d" type="ctc" on="no" exists="yes" pers="no" auto="no" failed="no" names=""
id="0.0.a18e:0.0.a18f" type="ctc" on="no" exists="yes" pers="no" auto="no" failed="no" names=""
id="0.0.a190:0.0.a191" type="ctc" on="no" exists="yes" pers="no" auto="no" failed="no" names=""
id="0.0.a192:0.0.a193" type="ctc" on="no" exists="yes" pers="no" auto="no" failed="no" names=""
id="0.0.a194:0.0.a195" type="ctc" on="no" exists="yes" pers="no" auto="no" failed="no" names=""
id="0.0.a196:0.0.a197" type="ctc" on="no" exists="yes" pers="no" auto="no" failed="no" names=""
id="0.0.a198:0.0.a199" type="ctc" on="no" exists="yes" pers="no" auto="no" failed="no" names=""
id="0.0.a19a:0.0.a19b" type="ctc" on="no" exists="yes" pers="no" auto="no" failed="no" names=""
id="0.0.a19c:0.0.a19d" type="ctc" on="no" exists="yes" pers="no" auto="no" failed="no" names=""
id="0.0.a19e:0.0.a19f" type="ctc" on="no" exists="yes" pers="no" auto="no" failed="no" names=""
id="0.0.a1a0:0.0.a1a1" type="ctc" on="no" exists="yes" pers="no" auto="no" failed="no" names=""
id="0.0.a1a2:0.0.a1a3" type="ctc" on="no" exists="yes" pers="no" auto="no" failed="no" names=""
id="0.0.a1a4:0.0.a1a5" type="ctc" on="no" exists="yes" pers="no" auto="no" failed="no" names=""
id="0.0.a1a6:0.0.a1a7" type="ctc" on="no" exists="yes" pers="no" auto="no" failed="no" names=""
id="0.0.a1a8:0.0.a1a9" type="ctc" on="no" exists="yes" pers="no" auto="no" failed="no" names=""
id="0.0.a1aa:0.0.a1ab" type="ctc" on="no" exists="yes" pers="no" auto="no" failed="no" names=""
id="0.0.a1ac:0.0.a1ad" type="ctc" on="no" exists="yes" pers="no" auto="no" failed="no" names=""
id="0.0.a1ae:0.0.a1af" type="ctc" on="no" exists="yes" pers="no" auto="no" failed="no" names=""
id="0.0.a1b0:0.0.a1b1" type="ctc" on="no" exists="yes" pers="no" auto="no" failed="no" names=""
id="0.0.a1b2:0.0.a1b3" type="ctc" on="no" exists="yes" pers="no" auto="no" failed="no" names=""
id="0.0.a1b4:0.0.a1b5" type="ctc" on="no" exists="yes" pers="no" auto="no" failed="no" names=""
id="0.0.a1b6:0.0.a1b7" type="ctc" on="no" exists="yes" pers="no" auto="no" failed="no" names=""
id="0.0.a1b8:0.0.a1b9" type="ctc" on="no" exists="yes" pers="no" auto="no" failed="no" names=""
id="0.0.a1ba:0.0.a1bb" type="ctc" on="no" exists="yes" pers="no" auto="no" failed="no" names=""
id="0.0.a1bc:0.0.a1bd" type="ctc" on="no" exists="yes" pers="no" auto="no" failed="no" names=""
id="0.0.a1be:0.0.a1bf" type="ctc" on="no" exists="yes" pers="no" auto="no" failed="no" names=""
id="0.0.a1c0:0.0.a1c1" type="ctc" on="no" exists="yes" pers="no" auto="no" failed="no" names=""
id="0.0.a1c2:0.0.a1c3" type="ctc" on="no" exists="yes" pers="no" auto="no" failed="no" names=""
id="0.0.a1c4:0.0.a1c5" type="ctc" on="no" exists="yes" pers="no" auto="no" failed="no" names=""
id="0.0.a1c6:0.0.a1c7" type="ctc" on="no" exists="yes" pers="no" auto="no" failed="no" names=""
id="0.0.a1c8:0.0.a1c9" type="ctc" on="no" exists="yes" pers="no" auto="no" failed="no" names=""
id="0.0.a1ca:0.0.a1cb" type="ctc" on="no" exists="yes" pers="no" auto="no" failed="no" names=""
id="0.0.a1cc:0.0.a1cd" type="ctc" on="no" exists="yes" pers="no" auto="no" failed="no" names=""
id="0.0.a1ce:0.0.a1cf" type="ctc" on="no" exists="yes" pers="no" auto="no" failed="no" names=""
id="0.0.a1d0:0.0.a1d1" type="ctc" on="no" exists="yes" pers="no" auto="no" failed="no" names=""
id="0.0.a1d2:0.0.a1d3" type="ctc" on="no" exists="yes" pers="no" auto="no" failed="no" names=""
id="0.0.a1d4:0.0.a1d5" type="ctc" on="no" exists="yes" pers="no" auto="no" failed="no" names=""
id="0.0.a1d6:0.0.a1d7" type="ctc" on="no" exists="yes" pers="no" auto="no" failed="no" names=""
id="0.0.a1d8:0.0.a1d9" type="ctc" on="no" exists="yes" pers="no" auto="no" failed="no" names=""
id="0.0.a1da:0.0.a1db" type="ctc" on="no" exists="yes" pers="no" auto="no" failed="no" names=""
id="0.0.a1dc:0.0.a1dd" type="ctc" on="no" exists="yes" pers="no" auto="no" failed="no" names=""
id="0.0.a1de:0.0.a1df" type="ctc" on="no" exists="yes" pers="no" auto="no" failed="no" names=""
id="0.0.a1e0:0.0.a1e1" type="ctc" on="no" exists="yes" pers="no" auto="no" failed="no" names=""
id="0.0.a1e2:0.0.a1e3" type="ctc" on="no" exists="yes" pers="no" auto="no" failed="no" names=""
id="0.0.a1e4:0.0.a1e5" type="ctc" on="no" exists="yes" pers="no" auto="no" failed="no" names=""
id="0.0.a1e6:0.0.a1e7" type="ctc" on="no" exists="yes" pers="no" auto="no" failed="no" names=""
id="0.0.a1e8:0.0.a1e9" type="ctc" on="no" exists="yes" pers="no" auto="no" failed="no" names=""
id="0.0.a1ea:0.0.a1eb" type="ctc" on="no" exists="yes" pers="no" auto="no" failed="no" names=""
id="0.0.a1ec:0.0.a1ed" type="ctc" on="no" exists="yes" pers="no" auto="no" failed="no" names=""
id="0.0.a1ee:0.0.a1ef" type="ctc" on="no" exists="yes" pers="no" auto="no" failed="no" names=""
id="0.0.a1f0:0.0.a1f1" type="ctc" on="no" exists="yes" pers="no" auto="no" failed="no" names=""
id="0.0.a1f2:0.0.a1f3" type="ctc" on="no" exists="yes" pers="no" auto="no" failed="no" names=""
id="0.0.a1f4:0.0.a1f5" type="ctc" on="no" exists="yes" pers="no" auto="no" failed="no" names=""
id="0.0.a1f6:0.0.a1f7" type="ctc" on="no" exists="yes" pers="no" auto="no" failed="no" names=""
id="0.0.a1f8:0.0.a1f9" type="ctc" on="no" exists="yes" pers="no" auto="no" failed="no" names=""
id="0.0.a1fa:0.0.a1fb" type="ctc" on="no" exists="yes" pers="no" auto="no" failed="no" names=""
id="0.0.a1fc:0.0.a1fd" type="ctc" on="no" exists="yes" pers="no" auto="no" failed="no" names=""
id="0.0.a1fe:0.0.a1ff" type="ctc" on="no" exists="yes" pers="no" auto="no" failed="no" names=""
id="0.0.f110:0.0.f111" type="ctc" on="no" exists="yes" pers="no" auto="no" failed="no" names=""
id="0.0.f112:0.0.f113" type="ctc" on="no" exists="yes" pers="no" auto="no" failed="no" names=""
id="0.0.f114:0.0.f115" type="ctc" on="no" exists="yes" pers="no" auto="no" failed="no" names=""
id="0.0.f116:0.0.f117" type="ctc" on="no" exists="yes" pers="no" auto="no" failed="no" names=""
id="0.0.f118:0.0.f119" type="ctc" on="no" exists="yes" pers="no" auto="no" failed="no" names=""
id="0.0.f11a:0.0.f11b" type="ctc" on="no" exists="yes" pers="no" auto="no" failed="no" names=""
id="0.0.f11c:0.0.f11d" type="ctc" on="no" exists="yes" pers="no" auto="no" failed="no" names=""
id="0.0.f11e:0.0.f11f" type="ctc" on="no" exists="yes" pers="no" auto="no" failed="no" names=""
id="0.0.f120:0.0.f121" type="ctc" on="no" exists="yes" pers="no" auto="no" failed="no" names=""
id="0.0.f122:0.0.f123" type="ctc" on="no" exists="yes" pers="no" auto="no" failed="no" names=""
id="0.0.f124:0.0.f125" type="ctc" on="no" exists="yes" pers="no" auto="no" failed="no" names=""
id="0.0.f126:0.0.f127" type="ctc" on="no" exists="yes" pers="no" auto="no" failed="no" names=""
id="0.0.f128:0.0.f129" type="ctc" on="no" exists="yes" pers="no" auto="no" failed="no" names=""
id="0.0.f12a:0.0.f12b" type="ctc" on="no" exists="yes" pers="no" auto="no" failed="no" names=""
id="0.0.f12c:0.0.f12d" type="ctc" on="no" exists="yes" pers="no" auto="no" failed="no" names=""
id="0.0.f12e:0.0.f12f" type="ctc" on="no" exists="yes" pers="no" auto="no" failed="no" names=""
id="0.0.f130:0.0.f131" type="ctc" on="no" exists="yes" pers="no" auto="no" failed="no" names=""
id="0.0.f132:0.0.f133" type="ctc" on="no" exists="yes" pers="no" auto="no" failed="no" names=""
id="0.0.f134:0.0.f135" type="ctc" on="no" exists="yes" pers="no" auto="no" failed="no" names=""
id="0.0.f136:0.0.f137" type="ctc" on="no" exists="yes" pers="no" auto="no" failed="no" names=""
id="0.0.f138:0.0.f139" type="ctc" on="no" exists="yes" pers="no" auto="no" failed="no" names=""
id="0.0.f13a:0.0.f13b" type="ctc" on="no" exists="yes" pers="no" auto="no" failed="no" names=""
id="0.0.f13c:0.0.f13d" type="ctc" on="no" exists="yes" pers="no" auto="no" failed="no" names=""
id="0.0.f13e:0.0.f13f" type="ctc" on="no" exists="yes" pers="no" auto="no" failed="no" names=""
id="0.0.f140:0.0.f141" type="ctc" on="no" exists="yes" pers="no" auto="no" failed="no" names=""
id="0.0.f142:0.0.f143" type="ctc" on="no" exists="yes" pers="no" auto="no" failed="no" names=""
id="0.0.f144:0.0.f145" type="ctc" on="no" exists="yes" pers="no" auto="no" failed="no" names=""
id="0.0.f146:0.0.f147" type="ctc" on="no" exists="yes" pers="no" auto="no" failed="no" names=""
id="0.0.f148:0.0.f149" type="ctc" on="no" exists="yes" pers="no" auto="no" failed="no" names=""
id="0.0.f14a:0.0.f14b" type="ctc" on="no" exists="yes" pers="no" auto="no" failed="no" names=""
id="0.0.f14c:0.0.f14d" type="ctc" on="no" exists="yes" pers="no" auto="no" failed="no" names=""
id="0.0.f14e:0.0.f14f" type="ctc" on="no" exists="yes" pers="no" auto="no" failed="no" names=""
id="0.0.f150:0.0.f151" type="ctc" on="no" exists="yes" pers="no" auto="no" failed="no" names=""
id="0.0.f152:0.0.f153" type="ctc" on="no" exists="yes" pers="no" auto="no" failed="no" names=""
id="0.0.f154:0.0.f155" type="ctc" on="no" exists="yes" pers="no" auto="no" failed="no" names=""
id="0.0.f156:0.0.f157" type="ctc" on="no" exists="yes" pers="no" auto="no" failed="no" names=""
id="0.0.f158:0.0.f159" type="ctc" on="no" exists="yes" pers="no" auto="no" failed="no" names=""
id="0.0.f15a:0.0.f15b" type="ctc" on="no" exists="yes" pers="no" auto="no" failed="no" names=""
id="0.0.f15c:0.0.f15d" type="ctc" on="no" exists="yes" pers="no" auto="no" failed="no" names=""
id="0.0.f15e:0.0.f15f" type="ctc" on="no" exists="yes" pers="no" auto="no" failed="no" names=""
id="0.0.f160:0.0.f161" type="ctc" on="no" exists="yes" pers="no" auto="no" failed="no" names=""
id="0.0.f162:0.0.f163" type="ctc" on="no" exists="yes" pers="no" auto="no" failed="no" names=""
id="0.0.f164:0.0.f165" type="ctc" on="no" exists="yes" pers="no" auto="no" failed="no" names=""
id="0.0.f166:0.0.f167" type="ctc" on="no" exists="yes" pers="no" auto="no" failed="no" names=""
id="0.0.f168:0.0.f169" type="ctc" on="no" exists="yes" pers="no" auto="no" failed="no" names=""
id="0.0.f16a:0.0.f16b" type="ctc" on="no" exists="yes" pers="no" auto="no" failed="no" names=""
id="0.0.f16c:0.0.f16d" type="ctc" on="no" exists="yes" pers="no" auto="no" failed="no" names=""
id="0.0.f16e:0.0.f16f" type="ctc" on="no" exists="yes" pers="no" auto="no" failed="no" names=""
id="0.0.f180:0.0.f181" type="ctc" on="no" exists="yes" pers="no" auto="no" failed="no" names=""
id="0.0.f182:0.0.f183" type="ctc" on="no" exists="yes" pers="no" auto="no" failed="no" names=""
id="0.0.f184:0.0.f185" type="ctc" on="no" exists="yes" pers="no" auto="no" failed="no" names=""
id="0.0.f186:0.0.f187" type="ctc" on="no" exists="yes" pers="no" auto="no" failed="no" names=""
id="0.0.f188:0.0.f189" type="ctc" on="no" exists="yes" pers="no" auto="no" failed="no" names=""
id="0.0.f18a:0.0.f18b" type="ctc" on="no" exists="yes" pers="no" auto="no" failed="no" names=""
id="0.0.f18c:0.0.f18d" type="ctc" on="no" exists="yes" pers="no" auto="no" failed="no" names=""
id="0.0.f18e:0.0.f18f" type="ctc" on="no" exists="yes" pers="no" auto="no" failed="no" names=""
id="0.0.f190:0.0.f191" type="ctc" on="no" exists="yes" pers="no" auto="no" failed="no" names=""
id="0.0.f192:0.0.f193" type="ctc" on="no" exists="yes" pers="no" auto="no" failed="no" names=""
id="0.0.f194:0.0.f195" type="ctc" on="no" exists="yes" pers="no" auto="no" failed="no" names=""
id="0.0.f196:0.0.f197" type="ctc" on="no" exists="yes" pers="no" auto="no" failed="no" names=""
id="0.0.f198:0.0.f199" type="ctc" on="no" exists="yes" pers="no" auto="no" failed="no" names=""
id="0.0.f19a:0.0.f19b" type="ctc" on="no" exists="yes" pers="no" auto="no" failed="no" names=""
id="0.0.f19c:0.0.f19d" type="ctc" on="no" exists="yes" pers="no" auto="no" failed="no" names=""
id="0.0.f19e:0.0.f19f" type="ctc" on="no" exists="yes" pers="no" auto="no" failed="no" names=""
id="0.0.f1a0:0.0.f1a1" type="ctc" on="no" exists="yes" pers="no" auto="no" failed="no" names=""
id="0.0.f1a2:0.0.f1a3" type="ctc" on="no" exists="yes" pers="no" auto="no" failed="no" names=""
id="0.0.f1a4:0.0.f1a5" type="ctc" on="no" exists="yes" pers="no" auto="no" failed="no" names=""
id="0.0.f1a6:0.0.f1a7" type="ctc" on="no" exists="yes" pers="no" auto="no" failed="no" names=""
id="0.0.f1a8:0.0.f1a9" type="ctc" on="no" exists="yes" pers="no" auto="no" failed="no" names=""
id="0.0.f1aa:0.0.f1ab" type="ctc" on="no" exists="yes" pers="no" auto="no" failed="no" names=""
id="0.0.f1ac:0.0.f1ad" type="ctc" on="no" exists="yes" pers="no" auto="no" failed="no" names=""
id="0.0.f1ae:0.0.f1af" type="ctc" on="no" exists="yes" pers="no" auto="no" failed="no" names=""
id="0.0.f1b0:0.0.f1b1" type="ctc" on="no" exists="yes" pers="no" auto="no" failed="no" names=""
id="0.0.f1b2:0.0.f1b3" type="ctc" on="no" exists="yes" pers="no" auto="no" failed="no" names=""
id="0.0.f1b4:0.0.f1b5" type="ctc" on="no" exists="yes" pers="no" auto="no" failed="no" names=""
id="0.0.f1b6:0.0.f1b7" type="ctc" on="no" exists="yes" pers="no" auto="no" failed="no" names=""
id="0.0.f1b8:0.0.f1b9" type="ctc" on="no" exists="yes" pers="no" auto="no" failed="no" names=""
id="0.0.f1ba:0.0.f1bb" type="ctc" on="no" exists="yes" pers="no" auto="no" failed="no" names=""
id="0.0.f1bc:0.0.f1bd" type="ctc" on="no" exists="yes" pers="no" auto="no" failed="no" names=""
id="0.0.f1be:0.0.f1bf" type="ctc" on="no" exists="yes" pers="no" auto="no" failed="no" names=""
id="0.0.f1c0:0.0.f1c1" type="ctc" on="no" exists="yes" pers="no" auto="no" failed="no" names=""
id="0.0.f1c2:0.0.f1c3" type="ctc" on="no" exists="yes" pers="no" auto="no" failed="no" names=""
id="0.0.f1c4:0.0.f1c5" type="ctc" on="no" exists="yes" pers="no" auto="no" failed="no" names=""
id="0.0.f1c6:0.0.f1c7" type="ctc" on="no" exists="yes" pers="no" auto="no" failed="no" names=""
id="0.0.f1c8:0.0.f1c9" type="ctc" on="no" exists="yes" pers="no" auto="no" failed="no" names=""
id="0.0.f1ca:0.0.f1cb" type="ctc" on="no" exists="yes" pers="no" auto="no" failed="no" names=""
id="0.0.f1cc:0.0.f1cd" type="ctc" on="no" exists="yes" pers="no" auto="no" failed="no" names=""
id="0.0.f1ce:0.0.f1cf" type="ctc" on="no" exists="yes" pers="no" auto="no" failed="no" names=""
id="0.0.f1d0:0.0.f1d1" type="ctc" on="no" exists="yes" pers="no" auto="no" failed="no" names=""
id="0.0.f1d2:0.0.f1d3" type="ctc" on="no" exists="yes" pers="no" auto="no" failed="no" names=""
id="0.0.f1d4:0.0.f1d5" type="ctc" on="no" exists="yes" pers="no" auto="no" failed="no" names=""
id="0.0.f1d6:0.0.f1d7" type="ctc" on="no" exists="yes" pers="no" auto="no" failed="no" names=""
id="0.0.f1d8:0.0.f1d9" type="ctc" on="no" exists="yes" pers="no" auto="no" failed="no" names=""
id="0.0.f1da:0.0.f1db" type="ctc" on="no" exists="yes" pers="no" auto="no" failed="no" names=""
id="0.0.f1dc:0.0.f1dd" type="ctc" on="no" exists="yes" pers="no" auto="no" failed="no" names=""
id="0.0.f1de:0.0.f1df" type="ctc" on="no" exists="yes" pers="no" auto="no" failed="no" names=""
id="0.0.f1e0:0.0.f1e1" type="ctc" on="no" exists="yes" pers="no" auto="no" failed="no" names=""
id="0.0.f1e2:0.0.f1e3" type="ctc" on="no" exists="yes" pers="no" auto="no" failed="no" names=""
id="0.0.f1e4:0.0.f1e5" type="ctc" on="no" exists="yes" pers="no" auto="no" failed="no" names=""
id="0.0.f1e6:0.0.f1e7" type="ctc" on="no" exists="yes" pers="no" auto="no" failed="no" names=""
id="0.0.f1e8:0.0.f1e9" type="ctc" on="no" exists="yes" pers="no" auto="no" failed="no" names=""
id="0.0.f1ea:0.0.f1eb" type="ctc" on="no" exists="yes" pers="no" auto="no" failed="no" names=""
id="0.0.f1ec:0.0.f1ed" type="ctc" on="no" exists="yes" pers="no" auto="no" failed="no" names=""
id="0.0.f1ee:0.0.f1ef" type="ctc" on="no" exists="yes" pers="no" auto="no" failed="no" names=""
id="0.0.f1f0:0.0.f1f1" type="ctc" on="no" exists="yes" pers="no" auto="no" failed="no" names=""
id="0.0.f1f2:0.0.f1f3" type="ctc" on="no" exists="yes" pers="no" auto="no" failed="no" names=""
id="0.0.f1f4:0.0.f1f5" type="ctc" on="no" exists="yes" pers="no" auto="no" failed="no" names=""
id="0.0.f1f6:0.0.f1f7" type="ctc" on="no" exists="yes" pers="no" auto="no" failed="no" names=""
id="0.0.f1f8:0.0.f1f9" type="ctc" on="no" exists="yes" pers="no" auto="no" failed="no" names=""
id="0.0.f1fa:0.0.f1fb" type="ctc" on="no" exists="yes" pers="no" auto="no" failed="no" names=""
id="0.0.f1fc:0.0.f1fd" type="ctc" on="no" exists="yes" pers="no" auto="no" failed="no" names=""
id="0.0.f1fe:0.0.f1ff" type="ctc" on="no" exists="yes" pers="no" auto="no" failed="no" names=""
id="0.0.c0fe" type="generic-ccw" on="no" exists="yes" pers="no" auto="no" failed="yes" names=""'''  # noqa: E501


@attr.s
class ZdevInfo:
    id = attr.ib()
    type = attr.ib()
    on = attr.ib()
    exists = attr.ib()
    pers = attr.ib()
    auto = attr.ib()
    failed = attr.ib()
    names = attr.ib()

    @classmethod
    def from_row(cls, row):
        row = dict((k.split('=', 1) for k in shlex.split(row)))
        for k, v in row.items():
            if v == "yes":
                row[k] = True
            if v == "no":
                row[k] = False
        return ZdevInfo(**row)

    @property
    def status(self):
        if self.failed:
            return Color.info_error(Text(_("failed"), align="center"))
        if self.auto and self.on:
            return Color.info_minor(Text(_("auto"), align="center"))
        if self.pers and self.on:
            return Text(_("online"), align="center")
        return Text("", align="center")

    @property
    def typeclass(self):
        if self.type.startswith('zfcp'):
            return 'zfcp'
        return self.type


class ZdevController(SubiquityController):

    def __init__(self, app):
        super().__init__(app)
        if self.opts.dry_run:
            if platform.machine() == 's390x':
                zdevinfos = self.lszdev()
            else:
                devices = lszdev_stock.splitlines()
                devices.sort()
                zdevinfos = [ZdevInfo.from_row(row) for row in devices]
            self.zdevinfos = OrderedDict([(i.id, i) for i in zdevinfos])

    def start_ui(self):
        if 'accept-default' in self.answers:
            self.done()
        self.ui.set_body(ZdevView(self))

    def cancel(self):
        self.app.prev_screen()

    def done(self):
        # switch to next screen
        self.app.next_screen()

    def chzdev(self, action, zdevinfo):
        if self.opts.dry_run:
            on = action == 'enable'
            self.zdevinfos[zdevinfo.id].on = on
            self.zdevinfos[zdevinfo.id].pers = on
        else:
            chzdev_cmd = ['chzdev', '--%s' % action, zdevinfo.id]
            run_command(chzdev_cmd)

    def get_zdevinfos(self):
        if self.opts.dry_run:
            return self.zdevinfos.values()
        else:
            return self.lszdev()

    def lszdev(self):
        devices = run_command(lszdev_cmd, universal_newlines=True).stdout
        devices = devices.splitlines()
        devices.sort()
        return [ZdevInfo.from_row(row) for row in devices]



# Copyright 2015 Canonical, Ltd.
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as
# published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import logging
import time
import yaml

from probert.network import (
    StoredDataObserver,
    UdevObserver,
    )

log = logging.getLogger('subiquitycore.prober')


class Prober():
    def __init__(self, machine_config, debug_flags):
        self.saved_config = None
        if machine_config:
            with open(machine_config) as mc:
                self.saved_config = yaml.safe_load(mc)
        self.debug_flags = debug_flags
        log.debug('Prober() init finished, data:{}'.format(self.saved_config))

    def probe_network(self, receiver):
        if self.saved_config is not None:
            observer = StoredDataObserver(
                self.saved_config['network'], receiver)
        else:
            observer = UdevObserver(receiver)
        return observer, observer.start()

    def get_storage(self, probe_types=None):
        if self.saved_config is not None:
            flag = 'bpfail-full'
            if probe_types is not None:
                flag = 'bpfail-restricted'
            if flag in self.debug_flags:
                time.sleep(2)
                1/0
            r = self.saved_config['storage'].copy()
            if probe_types is not None:
                for k in self.saved_config['storage']:
                    if k not in probe_types:
                        r[k] = {}
            return r
        from probert.storage import Storage
        return Storage().probe(probe_types=probe_types)



#!/usr/bin/env python3
# -*- mode: python; -*-
#
# Copyright 2015 Canonical, Ltd.
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This package is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.

"""
subiquity
=========
Ubuntu Server Installer
"""

from setuptools import setup, find_packages

import os
import sys

setup_kwargs = {}
# dpkg build uses build and install, tox uses sdist
if 'SUBIQUITY_NO_I18N' not in os.environ:
    from DistUtilsExtra.command import build_extra
    from DistUtilsExtra.command import build_i18n
    setup_kwargs['cmdclass'] = {'build': build_extra.build_extra,
                                'build_i18n': build_i18n.build_i18n}


with open(os.path.join(os.path.dirname(__file__),
                       'subiquitycore', '__init__.py')) as init:
    lines = [line for line in init if 'i18n' not in line]
    ns = {}
    exec('\n'.join(lines), ns)
    version = ns['__version__']

if sys.argv[-1] == 'clean':
    print("Cleaning up ...")
    os.system('rm -rf subiquity.egg-info build dist')
    sys.exit()

setup(name='subiquity',
      version=version,
      description="Ubuntu Server Installer",
      long_description=__doc__,
      author='Canonical Engineering',
      author_email='ubuntu-dev@lists.ubuntu.com',
      url='https://github.com/CanonicalLtd/subiquity',
      license="AGPLv3+",
      packages=find_packages(exclude=["tests"]),
      scripts=[
          'bin/console-conf-wait',
          'bin/console-conf-wrapper',
          'bin/subiquity-debug',
          'bin/subiquity-configure-apt',
          'bin/subiquity-configure-run',
          'bin/subiquity-loadkeys',
          'bin/subiquity-service',
      ],
      entry_points={
          'console_scripts': [
              'subiquity-tui = subiquity.cmd.tui:main',
              'console-conf-tui = console_conf.cmd.tui:main',
              ('console-conf-write-login-details = '
               'console_conf.cmd.write_login_details:main'),
          ],
      },
      data_files=[],
      **setup_kwargs)



# Copyright 2019 Canonical, Ltd.
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as
# published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import unittest

from subiquity.models.mirror import (
    MirrorModel,
    )


class TestMirrorModel(unittest.TestCase):

    def test_set_country(self):
        model = MirrorModel()
        model.set_country("CC")
        self.assertIn(
            model.get_mirror(),
            [
                "http://CC.archive.ubuntu.com/ubuntu",
                "http://CC.ports.ubuntu.com/ubuntu-ports",
            ])

    def test_set_mirror(self):
        model = MirrorModel()
        model.set_mirror("http://mymirror.invalid/")
        self.assertEqual(model.get_mirror(), "http://mymirror.invalid/")

    def test_set_country_after_set_mirror(self):
        model = MirrorModel()
        model.set_mirror("http://mymirror.invalid/")
        model.set_country("CC")
        self.assertEqual(model.get_mirror(), "http://mymirror.invalid/")



# Copyright 2015 Canonical, Ltd.
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as
# published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import logging
import os
import re

from urwid import (
    connect_signal,
    )

from subiquitycore.ui.interactive import (
    PasswordEditor,
    StringEditor,
    )
from subiquitycore.ui.form import (
    Form,
    simple_field,
    WantsToKnowFormField,
    )
from subiquitycore.ui.utils import screen
from subiquitycore.view import BaseView


log = logging.getLogger("subiquity.views.identity")

HOSTNAME_MAXLEN = 64
REALNAME_MAXLEN = 160
SSH_IMPORT_MAXLEN = 256 + 3  # account for lp: or gh:
USERNAME_MAXLEN = 32


class RealnameEditor(StringEditor, WantsToKnowFormField):
    def valid_char(self, ch):
        if len(ch) == 1 and ch in ':,=':
            self.bff.in_error = True
            self.bff.show_extra(("info_error",
                                 _("The characters : , and = are not permitted"
                                   " in this field")))
            return False
        else:
            return super().valid_char(ch)


class UsernameEditor(StringEditor, WantsToKnowFormField):
    def __init__(self):
        self.valid_char_pat = r'[-a-z0-9_]'
        self.error_invalid_char = _("The only characters permitted in this "
                                    "field are a-z, 0-9, _ and -")
        super().__init__()

    def valid_char(self, ch):
        if len(ch) == 1 and not re.match(self.valid_char_pat, ch):
            self.bff.in_error = True
            self.bff.show_extra(("info_error", self.error_invalid_char))
            return False
        else:
            return super().valid_char(ch)


RealnameField = simple_field(RealnameEditor)
UsernameField = simple_field(UsernameEditor)
PasswordField = simple_field(PasswordEditor)


class IdentityForm(Form):

    def __init__(self, reserved_usernames, initial):
        self.reserved_usernames = reserved_usernames
        super().__init__(initial=initial)

    realname = RealnameField(_("Your name:"))
    hostname = UsernameField(
        _("Your server's name:"),
        help=_("The name it uses when it talks to other computers."))
    username = UsernameField(_("Pick a username:"))
    password = PasswordField(_("Choose a password:"))
    confirm_password = PasswordField(_("Confirm your password:"))

    def validate_realname(self):
        if len(self.realname.value) > REALNAME_MAXLEN:
            return _("Realname too long, must be < ") + str(REALNAME_MAXLEN)

    def validate_hostname(self):
        if len(self.hostname.value) < 1:
            return _("Server name must not be empty")

        if len(self.hostname.value) > HOSTNAME_MAXLEN:
            return (_("Server name too long, must be < ") +
                    str(HOSTNAME_MAXLEN))

        if not re.match(r'[a-z_][a-z0-9_-]*', self.hostname.value):
            return _("Hostname must match NAME_REGEX, i.e. [a-z_][a-z0-9_-]*")

    def validate_username(self):
        username = self.username.value
        if len(username) < 1:
            return _("Username missing")

        if len(username) > USERNAME_MAXLEN:
            return _("Username too long, must be < ") + str(USERNAME_MAXLEN)

        if not re.match(r'[a-z_][a-z0-9_-]*', username):
            return _("Username must match NAME_REGEX, i.e. [a-z_][a-z0-9_-]*")

        if username in self.reserved_usernames:
            return _(
                'The username "{username}" is reserved for use by the system.'
                ).format(username=username)

    def validate_password(self):
        if len(self.password.value) < 1:
            return _("Password must be set")

    def validate_confirm_password(self):
        if self.password.value != self.confirm_password.value:
            return _("Passwords do not match")


def setup_password_validation(form, desc):
    def _check_password(sender, new_text):
        password = form.password.value
        if not password.startswith(new_text):
            form.confirm_password.show_extra(
                ("info_error", _("{desc} do not match").format(desc=desc)))
        else:
            form.confirm_password.show_extra('')
    connect_signal(
        form.confirm_password.widget, 'change', _check_password)


class IdentityView(BaseView):
    title = _("Profile setup")
    excerpt = _("Enter the username and password you will use to log in to "
                "the system. You can configure SSH access on the next screen "
                "but a password is still needed for sudo.")

    def __init__(self, model, controller):
        self.model = model
        self.controller = controller
        self.signal = controller.signal

        reserved_usernames_path = (
            os.path.join(os.environ.get("SNAP", "."), "reserved-usernames"))
        reserved_usernames = set()
        if os.path.exists(reserved_usernames_path):
            with open(reserved_usernames_path) as fp:
                for line in fp:
                    line = line.strip()
                    if line.startswith('#') or not line:
                        continue
                    reserved_usernames.add(line)
        else:
            reserved_usernames.add('root')

        if model.user:
            initial = {
                'realname': model.user.realname,
                'username': model.user.username,
                'hostname': model.hostname,
            }
        else:
            initial = {}

        self.form = IdentityForm(reserved_usernames, initial)

        connect_signal(self.form, 'submit', self.done)
        setup_password_validation(self.form, _("passwords"))

        super().__init__(
            screen(
                self.form.as_rows(),
                [self.form.done_btn],
                excerpt=_(self.excerpt),
                focus_buttons=False))

    def done(self, result):
        result = {
            "hostname": self.form.hostname.value,
            "realname": self.form.realname.value,
            "username": self.form.username.value,
            "password": self.model.encrypt_password(self.form.password.value),
        }
        self.controller.done(result)



# Copyright 2019 Canonical, Ltd.
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as
# published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import asyncio
import logging

from subiquitycore.async_helpers import schedule_task
from subiquitycore.controllers.network import NetworkController

from subiquity.controller import SubiquityController
from subiquity.controllers.error import ErrorReportKind


log = logging.getLogger("subiquity.controllers.network")


class NetworkController(NetworkController, SubiquityController):

    ai_data = None
    autoinstall_key = "network"
    autoinstall_schema = {
        'type': 'object',
        'properties': {
            'version': {
                'type': 'integer',
                'minimum': 2,
                'maximum': 2,
                },
            'ethernets': {'type': 'object'},
            'wifis': {'type': 'object'},
            'bridges': {'type': 'object'},
            'bonds': {'type': 'object'},
            'tunnels': {'type': 'object'},
            'vlans': {'type': 'object'},
            },
        }

    def __init__(self, app):
        super().__init__(app)
        app.note_file_for_apport("NetplanConfig", self.netplan_path)

    def load_autoinstall_data(self, data):
        self.ai_data = data

    def start(self):
        if self.ai_data is not None:
            self.apply_config()
        elif not self.interactive():
            self.initial_delay = schedule_task(self.delay())
        super().start()

    async def delay(self):
        await asyncio.sleep(10)

    async def apply_autoinstall_config(self):
        if self.ai_data is None:
            await self.initial_delay
            self.update_initial_configs()
            self.apply_config()
        await self.apply_config_task.wait()
        self.model.has_network = bool(
            self.network_event_receiver.default_routes)

    def render_config(self):
        if self.ai_data is not None:
            r = self.ai_data
            if self.interactive():
                # If we're interactive, we want later renders to
                # incorporate any changes from the UI.
                self.ai_data = None
            return r
        return super().render_config()

    async def _apply_config(self, silent):
        try:
            await super()._apply_config(silent)
        except asyncio.CancelledError:
            # asyncio.CancelledError is a subclass of Exception in
            # Python 3.6 (sadface)
            raise
        except Exception:
            log.exception("_apply_config failed")
            self.model.has_network = False
            self.app.make_apport_report(
                ErrorReportKind.NETWORK_FAIL, "applying network",
                interrupt=True)

    def done(self):
        self.configured()
        super().done()

    def make_autoinstall(self):
        return self.model.render_config()['network']



#!/usr/bin/python3

# This is just a helper script to generate lszdev commands
# later will use this more for "help" and description stuff

import subprocess
import shlex
columns=subprocess.check_output(['lszdev', '--pairs', '--list-columns'], universal_newlines=True)
rows = [shlex.split(i) for i in columns.splitlines()]
desired_columns=[]
for row in rows:
    record = dict([k.split('=',1) for k in row])
    if record['COLUMN'] not in ('TYPE', 'ATTR:', 'ATTRPATH:'):
        desired_columns.append(record['COLUMN'].lower())
cmd = ['lszdev','--pairs','--columns',','.join(desired_columns)]
print(cmd)
print(' '.join(cmd))



# Portions Copyright 2017 Canonical, Ltd.
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as
# published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

# This is adapted from
# https://github.com/pimutils/khal/commit/bd7c5f928a7670de9afae5657e66c6dc846688ac # noqa
#
#
# Copyright (c) 2013-2015 Christian Geier et al.
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to
# deal in the Software without restriction, including without limitation the
# rights to use, copy, modify, merge, publish, distribute, sublicense, and/or
# sell copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR OPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

"""Extended versions of urwid containers.

This module provides versions of urwid's Columns, Pile and ListBox
containers. It adds support for tab-cycling, callbacks on
gaining/losing focus and scrollbars on listboxes.

 1. Tab-cycling

Tab advances to the next selectable item and wraps when it gets to the
end of the screen. Shift tab goes the other way. In addition, an
unhandled "enter" key also advances to the next selectable item.

 2. Callbacks when a widget gains/loses focus

Define lost_focus or gained_focus callbacks on widgets and they'll be
called when the widget gains or loses focus.

 3. Scrollbars on ListBoxes.

All ListBoxes in subiquity gain a scrollbar when their contents don't
entirely fit on screen. The implementation assumes that there are not
too many elements in the ListBox.
"""

import logging
import operator

import urwid

log = logging.getLogger('subiquitycore.ui.container')

# The theory of tab-cycling:
#
# Tab should advance to the next selectable item and wrap when it gets
# to the end of the screen. Sounds simple, right? Well...
#
# Moving to the next selectable item within a container is pretty
# simple (it's what up and down do, after all). When a container is
# tabbed into, the first selectable within that container is selected
# and vv when it's shift tabbed into, the last selectable is
# selected. Urwid doesn't provide a native way to do that but it's
# pretty simple to implement (see the _select_first_selectable /
# _select_last_selectable methods). The tricky part is wrapping
# around: only the "outermost" container should wrap around. Urwid
# doesn't provide much of a way for a container to know about what it
# is contained in, so what this code does is, when interpreting a key
# that maps to either of the next selectable / prev selectable
# commands, passes a different key (the same key with " no wrap"
# appended) down to its contained widget. Then any containers within
# the outer one know not to wrap around in response to such a key.
#
# There is no support for a tab cycling Columns. It wouldn't be any
# harder than Pile but as subiquity is intended to be navigable with
# only up, down and enter, it doesn't actually contain any Columns
# with more than one selectable (this constraint is enforced by the
# Columns in this module).


def _maybe_call(w, methname):
    if w is None:
        return
    m = getattr(w.base_widget, methname, None)
    if m is not None:
        m()


def _has_other_selectable(widgets, cur_focus):
    for i, w in enumerate(widgets):
        if i != cur_focus and w.selectable():
            return True
    return False


for key, command in list(urwid.command_map._command.items()):
    if command in ('next selectable', 'prev selectable', urwid.ACTIVATE):
        urwid.command_map[key + ' no wrap'] = command

enter_advancing_command_map = urwid.command_map.copy()
enter_advancing_command_map['enter'] = 'next selectable'
enter_advancing_command_map['enter no wrap'] = 'next selectable'


class TabCyclingPile(urwid.Pile):
    _command_map = enter_advancing_command_map

    def selectable(self):
        for w, _ in self.contents:
            if w.selectable():
                return True
        return False

    def _select_first_selectable(self):
        """Select first selectable child (possibily recursively)."""
        for i, (w, o) in enumerate(self.contents):
            if w.selectable():
                self.set_focus(i)
                _maybe_call(w, "_select_first_selectable")
                return

    def _select_last_selectable(self):
        """Select last selectable child (possibily recursively)."""
        for i, (w, o) in reversed(list(enumerate(self.contents))):
            if w.selectable():
                self.set_focus(i)
                _maybe_call(w, "_select_last_selectable")
                return

    def _widgets(self):
        return [w for (w, o) in self._contents]

    # This is a copy/paste/hack of urwid's implementation. The main
    # reason we can't do this via simple subclassing is the need to
    # call _select_{first,last}_selectable on the newly focused
    # element when the focus changes.
    def keypress(self, size, key):
        """Pass the keypress to the widget in focus.
        Unhandled 'up' and 'down' keys may cause a focus change."""
        if not self.contents:
            return key

        # start subiquity change: new code
        downkey = key
        if not key.endswith(' no wrap') and (self._command_map[key] in
                                             ('next selectable',
                                              'prev selectable')):
            if _has_other_selectable(self._widgets(), self.focus_position):
                downkey += ' no wrap'
        # end subiquity change

        item_rows = None
        if len(size) == 2:
            item_rows = self.get_item_rows(size, focus=True)

        i = self.focus_position
        if self.selectable():
            tsize = self.get_item_size(size, i, True, item_rows)
            # start subiquity change: pass downkey to focus, not key,
            # do not return if command_map[upkey] is next/prev selectable
            upkey = self.focus.keypress(tsize, downkey)
            if self._command_map[upkey] not in ('cursor up', 'cursor down',
                                                'next selectable',
                                                'prev selectable'):
                return upkey
            # end subiquity change

        # begin subiquity change: new code
        if self._command_map[key] == 'next selectable':
            next_fp = self.focus_position + 1
            for i, (w, o) in enumerate(self._contents[next_fp:], next_fp):
                if w.selectable():
                    self.set_focus(i)
                    _maybe_call(w, "_select_first_selectable")
                    return
            if not key.endswith(' no wrap'):
                self._select_first_selectable()
            return key
        elif self._command_map[key] == 'prev selectable':
            positions = self._contents[:self.focus_position]
            for i, (w, o) in reversed(list(enumerate(positions))):
                if w.selectable():
                    self.set_focus(i)
                    _maybe_call(w, "_select_last_selectable")
                    return
            if not key.endswith(' no wrap'):
                self._select_last_selectable()
            return key
        # continued subiquity change: set 'meth' appropriately
        elif self._command_map[key] == 'cursor up':
            candidates = list(range(i-1, -1, -1))  # count backwards to 0
            meth = '_select_last_selectable'
        else:  # self._command_map[key] == 'cursor down'
            candidates = list(range(i+1, len(self.contents)))
            meth = '_select_first_selectable'
        # end subiquity change

        if not item_rows:
            item_rows = self.get_item_rows(size, focus=True)

        for j in candidates:
            if not self.contents[j][0].selectable():
                continue

            self._update_pref_col_from_focus(size)
            self.focus_position = j
            # begin subiquity change: new code
            _maybe_call(self.focus, meth)
            # end subiquity change
            if not hasattr(self.focus, 'move_cursor_to_coords'):
                return

            rows = item_rows[j]
            if self._command_map[key] == 'cursor up':
                rowlist = list(range(rows-1, -1, -1))
            else:  # self._command_map[key] == 'cursor down'
                rowlist = list(range(rows))
            for row in rowlist:
                tsize = self.get_item_size(size, j, True, item_rows)
                if self.focus_item.move_cursor_to_coords(
                        tsize, self.pref_col, row):
                    break
            return

        # nothing to select
        return key


class OneSelectableColumns(urwid.Columns):
    def __init__(self, widget_list, dividechars=0, focus_column=None,
                 min_width=1, box_columns=None):
        super().__init__(widget_list, dividechars, focus_column,
                         min_width, box_columns)
        selectables = 0
        for w, o in self._contents:
            if w.selectable():
                selectables += 1
        if selectables > 1:
            raise Exception(
                "subiquity only supports one selectable in a Columns")

    def _select_first_selectable(self):
        """Select first selectable child (possibily recursively)."""
        for i, (w, o) in enumerate(self.contents):
            if w.selectable():
                self.set_focus(i)
                _maybe_call(w, "_select_first_selectable")
                return

    def _select_last_selectable(self):
        """Select last selectable child (possibily recursively)."""
        for i, (w, o) in reversed(list(enumerate(self.contents))):
            if w.selectable():
                self.set_focus(i)
                _maybe_call(w, "_select_last_selectable")
                return


class TabCyclingListBox(urwid.ListBox):
    _command_map = enter_advancing_command_map

    def __init__(self, body):
        # urwid.ListBox converts an arbitrary sequence argument to a
        # PollingListWalker, which doesn't work with the below code.
        if getattr(body, 'get_focus', None) is None:
            body = urwid.SimpleFocusListWalker(body)
        super().__init__(body)

    def _set_focus_no_move(self, i):
        # We call set_focus twice because otherwise the listbox
        # attempts to do the minimal amount of scrolling required to
        # get the new focus widget into view, which is not what we
        # want, as if our first widget is a compound widget it results
        # its last widget being focused -- in fact the opposite of
        # what we want!
        self.set_focus(i)
        self.set_focus(i)
        # I don't really understand why this is required but it seems it is.
        self._invalidate()

    def _select_first_selectable(self):
        """Select first selectable child (possibily recursively)."""
        for i, w in enumerate(self.body):
            if w.selectable():
                self._set_focus_no_move(i)
                _maybe_call(w, "_select_first_selectable")
                return

    def _select_last_selectable(self):
        """Select last selectable child (possibily recursively)."""
        for i, w in reversed(list(enumerate(self.body))):
            if w.selectable():
                self._set_focus_no_move(i)
                _maybe_call(w, "_select_last_selectable")
                return

    def keypress(self, size, key):
        downkey = key
        if not key.endswith(' no wrap') and (self._command_map[key] in
                                             ('next selectable',
                                              'prev selectable')):
            if _has_other_selectable(self.body, self.focus_position):
                downkey += ' no wrap'
        upkey = super().keypress(size, downkey)
        if upkey != downkey:
            key = upkey

        if len(self.body) == 0:
            return key

        if self._command_map[key] == 'next selectable':
            next_fp = self.focus_position + 1
            for i, w in enumerate(self.body[next_fp:], next_fp):
                if w.selectable():
                    self.set_focus(i)
                    _maybe_call(w, "_select_first_selectable")
                    return None
            if not key.endswith(' no wrap'):
                self._select_first_selectable()
            return key
        elif self._command_map[key] == 'prev selectable':
            positions = self.body[:self.focus_position]
            for i, w in reversed(list(enumerate(positions))):
                if w.selectable():
                    self.set_focus(i)
                    _maybe_call(w, "_select_last_selectable")
                    return None
            if not key.endswith(' no wrap'):
                self._select_last_selectable()
            return key
        else:
            return key


class FocusTrackingMixin:

    def __init__(self, *args, **kw):
        super().__init__(*args, **kw)
        self._contents.set_focus_changed_callback(self._focus_changed)

    def lost_focus(self):
        try:
            cur_focus = self.focus
        except IndexError:
            pass
        else:
            _maybe_call(cur_focus, 'lost_focus')

    def gained_focus(self):
        _maybe_call(self.focus, 'gained_focus')

    def _focus_changed(self, new_focus):
        try:
            cur_focus = self.focus
        except IndexError:
            pass
        else:
            _maybe_call(cur_focus, 'lost_focus')
        _maybe_call(self[new_focus], 'gained_focus')
        self._invalidate()


class FocusTrackingPile(FocusTrackingMixin, TabCyclingPile):
    pass


class FocusTrackingColumns(FocusTrackingMixin, OneSelectableColumns):
    pass


class FocusTrackingListBox(TabCyclingListBox):

    def __init__(self, *args, **kw):
        super().__init__(*args, **kw)
        self.body.set_focus_changed_callback(self._focus_changed)
        self._in_set_focus_complete = False

    def _focus_changed(self, new_focus):
        if not self._in_set_focus_complete:
            _maybe_call(self.focus, 'lost_focus')
            _maybe_call(self.body[new_focus], 'gained_focus')
            self._invalidate()

    def lost_focus(self):
        _maybe_call(self.focus, 'lost_focus')

    def gained_focus(self):
        _maybe_call(self.focus, 'gained_focus')

    def _set_focus_complete(self, size, focus):
        self._in_set_focus_complete = True
        try:
            super()._set_focus_complete(size, focus)
        finally:
            self._in_set_focus_complete = False


Columns = FocusTrackingColumns
Pile = FocusTrackingPile


class ScrollBarListBox(urwid.WidgetDecoration):

    top = urwid.SolidFill("\N{BLACK UP-POINTING SMALL TRIANGLE}")
    box = urwid.SolidFill("\N{FULL BLOCK}")
    bg = urwid.SolidFill(" ")
    bot = urwid.SolidFill("\N{BLACK DOWN-POINTING SMALL TRIANGLE}")

    def __init__(self, lb, *, always_scroll=False):
        pile = Pile([
                ('fixed',  1, self.top),
                ('weight', 1, self.bg),
                ('weight', 1, self.box),
                ('weight', 1, self.bg),
                ('fixed',  1, self.bot),
            ])
        self.bar = urwid.AttrMap(pile, 'scrollbar', 'scrollbar focus')
        self._always_scroll = always_scroll
        super().__init__(lb)

    def _scroll(self, size, focus):
        if self._always_scroll:
            return True
        return len(self.original_widget.ends_visible(size, focus)) == 2

    def keypress(self, size, key):
        lb = self.original_widget
        if not self._scroll(size, True):
            size = (size[0]-1, size[1])
        return lb.keypress(size, key)

    def render(self, size, focus=False):
        lb = self.original_widget
        if self._scroll(size, focus):
            return lb.render(size, focus)
        else:
            # This implementation assumes that the number of rows is
            # not too large (and in particular is finite). That's the
            # case for all the listboxes we have in subiquity today.
            maxcol, maxrow = size

            offset, inset = lb.get_focus_offset_inset((maxcol - 1, maxrow))
            visible = lb.ends_visible((maxcol - 1, maxrow), focus)

            seen_focus = False
            height = height_before_focus = 0
            focus_widget, focus_pos = lb.body.get_focus()
            # Scan through the rows calculating total height and the
            # height of the rows before the focus widget.
            for widget in lb.body:
                rows = widget.rows((maxcol - 1,))
                if widget is focus_widget:
                    seen_focus = True
                elif not seen_focus:
                    height_before_focus += rows
                height += rows

            # Calculate the number of rows off the top and bottom of
            # the listbox.
            if 'top' in visible:
                top = 0
            else:
                top = height_before_focus + inset - offset
            if 'bottom' in visible:
                bottom = 0
            else:
                bottom = height - top - maxrow

            # Prevent the box from being squished to 0 rows. Input of
            #
            #     ('weight', maxrow)
            #
            # gets
            #
            #     round(maxrow / (top + maxrow + bottom))
            #
            # rows and that being < 1 simplifies to the below.
            o = self.bar.base_widget.options
            if maxrow < top + bottom:
                boxopt = o('given', 1)
            else:
                boxopt = o('weight', maxrow)

            top_arrow = self.top
            if top == 0:
                top_arrow = urwid.AttrMap(top_arrow, 'scrollbar')
            bot_arrow = self.bot
            if bottom == 0:
                bot_arrow = urwid.AttrMap(bot_arrow, 'scrollbar')

            self.bar.base_widget.contents[:] = [
                (top_arrow, o('given', 1)),
                (self.bg,  o('weight', top)),
                (self.box, boxopt),
                (self.bg,  o('weight', bottom)),
                (bot_arrow, o('given', 1)),
                ]
            lb_canvas = lb.render((maxcol - 1, maxrow), focus)
            bar_canvas = self.bar.render((1, maxrow), focus)
            return urwid.CanvasJoin([
                (lb_canvas,  lb.focus_position, True,  maxcol - 1),
                (bar_canvas, None,              False, 1),
            ])


def ListBox(body=None, *, always_scroll=False):
    # urwid.ListBox converts an arbitrary sequence argument to a
    # PollingListWalker, which doesn't work with our code.
    if body is None:
        body = []
    if body is getattr(body, 'get_focus', None) is None:
        body = urwid.SimpleFocusListWalker(body)
    return ScrollBarListBox(
        FocusTrackingListBox(body),
        always_scroll=always_scroll)


get_delegate = operator.attrgetter("_wrapped_widget.base_widget")


class OurWidgetWrap(urwid.WidgetWrap):
    # A wrapped widget needs to have a _select_first/last_selectable
    # method if and only if the widget being wrapped does, so we
    # define our own WidgetWrap class that uses the same technique to
    # forward these methods along if present as urwid's WidgetWrap
    # does.

    _select_first_selectable = property(
        lambda self: get_delegate(self)._select_first_selectable)
    _select_last_selectable = property(
        lambda self: get_delegate(self)._select_last_selectable)
    lost_focus = property(lambda self: get_delegate(self).lost_focus)
    gained_focus = property(lambda self: get_delegate(self).gained_focus)


WidgetWrap = OurWidgetWrap



# Copyright 2015 Canonical, Ltd.
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as
# published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

""" ConsoleConf UI Views """

from .identity import IdentityView
from .login import LoginView
from .welcome import WelcomeView, ChooserWelcomeView
from .chooser import (
    ChooserView,
    ChooserCurrentSystemView,
    ChooserConfirmView
    )

__all__ = [
    "IdentityView",
    "LoginView",
    "WelcomeView",
    "ChooserWelcomeView",
    "ChooserView",
    "ChooserCurrentSystemView",
    "ChooserConfirmView",
]



import unittest
from unittest import mock

import urwid

from subiquitycore.testing import view_helpers

from subiquity.controllers.filesystem import FilesystemController
from subiquity.models.filesystem import (
    Bootloader,
    Disk,
    FilesystemModel,
    )
from subiquity.models.tests.test_filesystem import (
    FakeStorageInfo,
    )
from subiquity.ui.views.filesystem.filesystem import FilesystemView


class FilesystemViewTests(unittest.TestCase):

    def make_view(self, model, devices=[]):
        controller = mock.create_autospec(spec=FilesystemController)
        controller.ui = mock.Mock()
        model.bootloader = Bootloader.NONE
        model.all_devices.return_value = devices
        model.grub_install_device = None
        return FilesystemView(model, controller)

    def test_simple(self):
        self.make_view(mock.create_autospec(spec=FilesystemModel))

    def test_one_disk(self):
        model = mock.create_autospec(spec=FilesystemModel)
        model._orig_config = []
        model._actions = []
        disk = Disk(
            m=model, serial="DISK-SERIAL", path='/dev/thing',
            info=FakeStorageInfo(size=100*(2**20), free=50*(2**20)))
        view = self.make_view(model, [disk])
        w = view_helpers.find_with_pred(
            view,
            lambda w: isinstance(w, urwid.Text) and "DISK-SERIAL" in w.text)
        self.assertIsNotNone(w, "could not find DISK-SERIAL in view")
